import math
from collections import defaultdict
from typing import Callable, Any
from uuid import uuid4
from ast import literal_eval

import networkx as nx
import numpy as np
import pandas as pd

from sampo.schemas import Time
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode, WorkGraph, EdgeType
from sampo.schemas.requirements import WorkerReq, ZoneReq
from sampo.schemas.resources import Worker
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.works import WorkUnit
from sampo.utilities.name_mapper import NameMapper

UNKNOWN_CONN_TYPE = 0
NONE_ELEM = '-1'


class Graph:
    def __init__(self):
        self.graph = defaultdict(list)

    def add_edge(self, u, v, weight=None):
        self.graph[u].append((v, weight))

    def dfs_cycle(self, u, visited):
        stack = [u]

        while len(stack) > 0:
            v = stack.pop()
            if visited[v] == 1:
                visited[v] = 2
                continue
            if visited[v] == 2:
                continue
            visited[v] = 1
            stack.append(v)
            for neighbour, weight in self.graph[v]:
                if visited[neighbour] == 0:
                    stack.append(neighbour)
                elif visited[neighbour] == 1:
                    return True
        return False

    def find_cycle(self):
        visited = {}
        for key, value in self.graph.items():
            for v in value:
                visited[v[0]] = 0
            visited[key] = 0
        path = []

        for key, value in visited.items():
            if value == 0:
                if self.dfs_cycle(key, visited):
                    v = key
                    path.append(v)
                    while v not in path[:-1]:
                        for n, _ in self.graph[v]:
                            if visited[n] == 1:
                                v = n
                                break
                        path.append(v)
                    for v in path:
                        visited[v] = 2
                    return path

        return None

    def eliminate_cycle(self, cycle):
        min_weight = float('inf')
        min_edge = None

        for i in range(len(cycle) - 1):
            u = cycle[i]
            v = cycle[i + 1]

            for neighbor, weight in self.graph[u]:
                if weight is None:
                    weight = 1
                if neighbor == v and weight <= min_weight:
                    min_weight = weight
                    min_edge = (u, v)

        u, v = min_edge
        self.graph[u] = [(neighbor, weight) for neighbor, weight in self.graph[u] if neighbor != v]

    def eliminate_cycles(self, is_eliminate_cycle: bool = True) -> list | None:
        cycle = self.find_cycle()
        cycles = []

        while cycle is not None:
            for i in range(len(cycle)):
                if cycle[i] == cycle[-1]:
                    cycle = cycle[i:]
                    break
            if not is_eliminate_cycle:
                cycles.append(cycle)
                self.eliminate_cycle(cycle)
            else:
                self.eliminate_cycle(cycle)
            cycle = self.find_cycle()
        if not is_eliminate_cycle:
            if cycles:
                return cycles
            return None
        for v in self.graph:
            self.graph[v] = [u for u, w in self.graph[v]]


def break_loops_in_input_graph(works_info: pd.DataFrame) -> pd.DataFrame:
    """
    Eliminate all cycles in received work graph. Algo breaks cycle by removing edge with the lowest weight
    (e.x. frequency of occurrence of the link in historical data).

    :param works_info: given work info
    :return: work info without cycles
    """
    graph = Graph()
    for _, row in works_info.iterrows():
        for pred_id, con_type, lag, counts in zip(row['predecessor_ids'], row['connection_types'], row['lags'],
                                                  row['counts']):
            if pred_id == '-1':
                continue
            graph.add_edge(pred_id, row['activity_id'], counts)

    graph.eliminate_cycles()
    for _, row in works_info.iterrows():
        i = 0
        while i < len(row['predecessor_ids']):
            if row['predecessor_ids'][i] not in graph.graph:
                i += 1
                continue
            if row['activity_id'] in graph.graph[row['predecessor_ids'][i]]:
                i += 1
                continue
            del row['predecessor_ids'][i]
            del row['connection_types'][i]
            del row['lags'][i]
            del row['counts'][i]
    return works_info.drop(columns=['counts'])


def fix_df_column_with_arrays(column: pd.Series, cast: Callable[[str], Any] | None = str,
                              none_elem: Any | None = NONE_ELEM) -> pd.Series:
    new_column = column.copy().astype(str).apply(
        lambda elems: [cast(elem) if elem != '' and elem != str(math.nan) else none_elem for elem in
                       elems.split(',')] if (elems != str(math.nan) and elems.split(',') != '') else [none_elem])
    return new_column


def preprocess_graph_df(frame: pd.DataFrame,
                        name_mapper: NameMapper | None = None) -> pd.DataFrame:
    def normalize_if_number(s):
        return str(int(float(s))) \
            if s.replace('.', '', 1).isdigit() \
            else s

    temp_lst = [math.nan] * frame.shape[0]

    for col in ['predecessor_ids', 'connection_types', 'lags', 'counts']:
        if col not in frame.columns:
            frame[col] = temp_lst

    if 'granular_name' not in frame.columns:
        frame['granular_name'] = [name_mapper[activity_name] for activity_name in frame['activity_name']]

    frame['activity_id'] = frame['activity_id'].astype(str)
    frame['volume'] = [float(x.replace(',', '.')) if isinstance(x, str) else float(x) for x in frame['volume']]

    if 'min_req' in frame.columns and 'max_req' in frame.columns:
        frame['min_req'] = [literal_eval(x) if isinstance(x, str) else x for x in frame['min_req']]
        frame['max_req'] = [literal_eval(x) if isinstance(x, str) else x for x in frame['max_req']]

    frame['predecessor_ids'] = fix_df_column_with_arrays(frame['predecessor_ids'], cast=normalize_if_number)
    frame['connection_types'] = fix_df_column_with_arrays(frame['connection_types'],
                                                          cast=EdgeType,
                                                          none_elem=EdgeType.FinishStart)
    if 'lags' not in frame.columns:
        frame['lags'] = [NONE_ELEM] * len(frame)
    frame['lags'] = fix_df_column_with_arrays(frame['lags'], float)
    for col in ['predecessor_ids', 'connection_types', 'lags', 'counts']:
        frame[col] = frame[col].astype(object)

    for _, row in frame.iterrows():
        frame.at[_, 'counts'] = [np.iinfo(np.int64).max] * len(frame.at[_, 'lags'])

    return frame


def add_graph_info(frame: pd.DataFrame) -> pd.DataFrame:
    existed_ids = set(frame['activity_id'])

    predecessor_ids, connection_types, lags = [], [], []
    for _, row in frame[['predecessor_ids', 'connection_types', 'lags']].iterrows():
        predecessor_ids.append([])
        connection_types.append([])
        lags.append([])
        for index in range(len(row['predecessor_ids'])):
            if row['predecessor_ids'][index] in existed_ids:
                predecessor_ids[-1].append(row['predecessor_ids'][index])
                connection_types[-1].append(row['connection_types'][index])
                lags[-1].append(row['lags'][index])
    frame['predecessor_ids'], frame['connection_types'], frame['lags'] = predecessor_ids, connection_types, lags

    frame['edges'] = frame[['predecessor_ids', 'connection_types', 'lags']].apply(lambda row: list(zip(*row)), axis=1)
    return frame


def topsort_graph_df(frame: pd.DataFrame) -> pd.DataFrame:
    """
    Sort DataFrame of works in topological order

    :param frame: DataFrame of works
    :return: topologically sorted DataFrame
    """
    G = nx.DiGraph()
    for _, row in frame.iterrows():
        G.add_node(row['activity_id'])
        for pred_id, con_type, lag in row['edges']:
            G.add_edge(pred_id, row['activity_id'])

    sorted_nodes = list(nx.topological_sort(G))
    frame['sort_key'] = frame['activity_id'].apply(lambda x: sorted_nodes.index(x))
    frame = frame.sort_values('sort_key')

    return frame


def build_work_graph(frame: pd.DataFrame, resource_names: list[str], work_estimator: WorkTimeEstimator) -> WorkGraph:
    id_to_node = {}

    for _, row in frame.iterrows():
        if 'min_req' in frame.columns and 'max_req' in frame.columns and 'req_volume' in frame.columns:
            reqs = []
            for res_name in resource_names:
                if res_name in row['min_req'] and res_name in row['max_req']:
                    if 0 < row['min_req'][res_name] <= row['max_req'][res_name]:
                        reqs.append(WorkerReq(res_name, Time(int(row['req_volume'][res_name])),
                                              row['min_req'][res_name],
                                              row['max_req'][res_name]))
        else:
            reqs = work_estimator.find_work_resources(work_name=row['granular_name'],
                                                      work_volume=float(row['volume']),
                                                      measurement=row['measurement'])
        is_service_unit = len(reqs) == 0

        zone_reqs = [ZoneReq(*v) for v in eval(row['required_statuses']).items()] \
            if 'required_statuses' in frame.columns else []

        description = row['description'] if 'description' in frame.columns else ''
        group = row['group'] if 'group' in frame.columns else 'main project'
        priority = row['priority'] if 'priority' in frame.columns else 1

        work_unit = WorkUnit(row['activity_id'], row['granular_name'], reqs, group=group,
                             description=description, volume=row['volume'], volume_type=row['measurement'],
                             is_service_unit=is_service_unit, display_name=row['activity_name_original'],
                             zone_reqs=zone_reqs, priority=priority)
        parents = [(id_to_node[p_id], lag, conn_type) for p_id, conn_type, lag in row.edges]
        node = GraphNode(work_unit, parents)
        id_to_node[row['activity_id']] = node

    all_nodes = [id_to_node[index] for index in list(set(id_to_node.keys()))]
    return WorkGraph.from_nodes(all_nodes)


def get_graph_contractors(path: str, contractor_name: str | None = 'ООО "***"') -> (
        list[Contractor], dict[str, float]):
    contractor_id = str(uuid4())
    workers_df = pd.read_csv(path, index_col='name')
    workers = {(name, 0): Worker(str(uuid4()), name, count * 2, contractor_id)
               for name, count in zip(workers_df.index, workers_df['count'])}
    workers_max_count = workers_df['count'].to_dict()
    contractors = [Contractor(contractor_id, contractor_name, list(workers_max_count.keys()), [], workers, {})]
    return contractors, workers_max_count



import math
from typing import Callable

import sampo.scheduler

from random import Random

import pathos.multiprocessing

from sampo.api.genetic_api import ChromosomeType, FitnessFunction, ScheduleGenerationScheme
from sampo.backend import T, R
from sampo.backend.default import DefaultComputationalBackend
from sampo.scheduler.genetic.operators import Individual
from sampo.scheduler.genetic.utils import create_toolbox_using_cached_chromosomes, init_chromosomes_f
from sampo.scheduler.heft import HEFTScheduler, HEFTBetweenScheduler
from sampo.scheduler.resource import AverageReqResourceOptimizer
from sampo.scheduler.resources_in_time import AverageBinarySearchResourceOptimizingScheduler
from sampo.scheduler.topological.base import RandomizedTopologicalScheduler
from sampo.schemas import WorkGraph, Contractor, LandscapeConfiguration, Time, WorkTimeEstimator, Schedule, GraphNode, \
    NoSufficientContractorError
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time_estimator import DefaultWorkEstimator


def scheduler_info_initializer(wg: WorkGraph,
                               contractors: list[Contractor],
                               landscape: LandscapeConfiguration,
                               spec: ScheduleSpec,
                               selection_size: int,
                               mutate_order: float,
                               mutate_resources: float,
                               mutate_zones: float,
                               deadline: Time | None,
                               weights: list[int] | None,
                               init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]],
                               assigned_parent_time: Time,
                               fitness_weights: tuple[int | float, ...],
                               rand: Random | None,
                               work_estimator_recreate_params: tuple | None,
                               sgs_type: ScheduleGenerationScheme,
                               only_lft_initialization: bool,
                               is_multiobjective: bool):
    global g_wg, g_contractors, g_landscape, g_spec, g_toolbox, g_work_estimator, g_deadline, g_rand, g_weights, \
        g_sgs_type, g_only_lft_initialization, g_is_multiobjective

    g_wg = wg
    g_contractors = contractors
    g_landscape = landscape
    g_spec = spec
    g_deadline = deadline
    g_weights = weights
    g_rand = rand
    g_sgs_type = sgs_type
    g_only_lft_initialization = only_lft_initialization
    g_is_multiobjective = is_multiobjective

    assigned_parent_time = assigned_parent_time or Time(0)

    g_work_estimator = work_estimator_recreate_params[0](*work_estimator_recreate_params[1])

    if init_chromosomes is not None:
        g_toolbox = create_toolbox_using_cached_chromosomes(wg,
                                                            contractors,
                                                            selection_size,
                                                            mutate_order,
                                                            mutate_resources,
                                                            mutate_zones,
                                                            init_chromosomes,
                                                            rand,
                                                            spec,
                                                            g_work_estimator,
                                                            assigned_parent_time,
                                                            fitness_weights,
                                                            landscape,
                                                            sgs_type,
                                                            only_lft_initialization,
                                                            is_multiobjective)


class MultiprocessingComputationalBackend(DefaultComputationalBackend):

    def __init__(self, n_cpus: int):
        self._n_cpus = n_cpus
        self._init_chromosomes = None
        self._pool = None
        super().__init__()

    def map(self, action: Callable[[T], R], values: list[T]) -> list[R]:
        return self._pool.map(action, values)

    def _ensure_pool_created(self):
        if self._pool is not None:
            return
        self._pool = pathos.multiprocessing.Pool(self._n_cpus,
                                                 initializer=scheduler_info_initializer,
                                                 initargs=(self._wg,
                                                           self._contractors,
                                                           self._landscape,
                                                           self._spec,
                                                           self._selection_size,
                                                           self._mutate_order,
                                                           self._mutate_resources,
                                                           self._mutate_zones,
                                                           self._deadline,
                                                           self._weights,
                                                           self._init_chromosomes,
                                                           self._assigned_parent_time,
                                                           self._fitness_weights,
                                                           self._rand,
                                                           self._work_estimator.get_recreate_info(),
                                                           self._sgs_type,
                                                           self._only_lft_initialization,
                                                           self._is_multiobjective))

    def cache_scheduler_info(self,
                             wg: WorkGraph,
                             contractors: list[Contractor],
                             landscape: LandscapeConfiguration = LandscapeConfiguration(),
                             spec: ScheduleSpec = ScheduleSpec(),
                             rand: Random | None = None,
                             work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().cache_scheduler_info(wg, contractors, landscape, spec, rand, work_estimator)
        self._pool = None

    def cache_genetic_info(self,
                           population_size: int = 50,
                           mutate_order: float = 0.1,
                           mutate_resources: float = 0.05,
                           mutate_zones: float = 0.05,
                           deadline: Time | None = None,
                           weights: list[int] | None = None,
                           init_schedules: dict[
                               str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]] = None,
                           assigned_parent_time: Time = Time(0),
                           fitness_weights: tuple[int | float, ...] = None,
                           sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                           only_lft_initialization: bool = False,
                           is_multiobjective: bool = False):
        super().cache_genetic_info(population_size, mutate_order, mutate_resources, mutate_zones, deadline,
                                   weights, init_schedules, assigned_parent_time, fitness_weights, sgs_type,
                                   only_lft_initialization, is_multiobjective)
        if init_schedules:
            self._init_chromosomes = init_chromosomes_f(self._wg, self._contractors, init_schedules,
                                                  self._landscape)
        else:
            self._init_chromosomes = []
        self._pool = None

    def compute_chromosomes(self, fitness: FitnessFunction, chromosomes: list[ChromosomeType]) -> list[float]:
        self._ensure_pool_created()

        def mapper(chromosome):
            return fitness.evaluate(chromosome, g_toolbox.evaluate_chromosome)

        return self.map(mapper, chromosomes)

    def generate_first_population(self, size_population: int) -> list[Individual]:
        self._ensure_toolbox_created()
        self._ensure_pool_created()

        def mapper(key: str):
            def randomized_init():
                schedule, _, _, order = RandomizedTopologicalScheduler(g_work_estimator, int(g_rand.random() * 1000000)) \
                    .schedule_with_cache(g_wg, g_contractors, landscape=g_landscape)[0]
                return schedule, order, g_spec

            def init_k_schedule(scheduler_class, k) -> tuple[Schedule | None, list[GraphNode] | None, ScheduleSpec | None]:
                try:
                    schedule, _, _, node_order = (scheduler_class(work_estimator=g_work_estimator,
                                                                  resource_optimizer=AverageReqResourceOptimizer(k))
                                                  .schedule_with_cache(g_wg, g_contractors, g_spec,
                                                                       landscape=g_landscape))[0]
                    return schedule, node_order[::-1], g_spec
                except NoSufficientContractorError:
                    return None, None, None

            if g_deadline is None:
                def init_schedule(scheduler_class) -> tuple[Schedule | None, list[GraphNode] | None, ScheduleSpec | None]:
                    try:
                        schedule, _, _, node_order = (scheduler_class(work_estimator=g_work_estimator)
                                                      .schedule_with_cache(g_wg, g_contractors, g_spec,
                                                                           landscape=g_landscape))[0]
                        return schedule, node_order[::-1], g_spec
                    except NoSufficientContractorError:
                        return None, None, None

            else:
                def init_schedule(scheduler_class) -> tuple[Schedule | None, list[GraphNode] | None, ScheduleSpec | None]:
                    try:
                        (schedule, _, _, node_order), modified_spec = AverageBinarySearchResourceOptimizingScheduler(
                            scheduler_class(work_estimator=g_work_estimator)
                        ).schedule_with_cache(g_wg, g_contractors, g_deadline, g_spec, landscape=g_landscape)
                        return schedule, node_order[::-1], modified_spec
                    except NoSufficientContractorError:
                        return None, None, None

            def convert(schedule: Schedule, priority_list: list[GraphNode], spec: ScheduleSpec):
                return g_toolbox.schedule_to_chromosome(schedule=schedule, spec=spec, order=priority_list)

            match key:
                case 'heft_end':
                    return convert(*init_schedule(HEFTScheduler))
                case 'heft_between':
                    return convert(*init_schedule(HEFTBetweenScheduler))
                case '12.5%':
                    return convert(*init_k_schedule(HEFTScheduler, 8))
                case '25%':
                    return convert(*init_k_schedule(HEFTScheduler, 4))
                case '75%':
                    return convert(*init_k_schedule(HEFTScheduler, 4 / 3))
                case '87.5%':
                    return convert(*init_k_schedule(HEFTScheduler, 8 / 7))
                case 'randomized':
                    return convert(*randomized_init())

        weights = self._weights or [2, 2, 1, 1, 1, 1]

        count_for_specified_types = (size_population // 3) // len(weights)
        count_for_specified_types = count_for_specified_types if count_for_specified_types > 0 else 1
        sum_counts_for_specified_types = count_for_specified_types * len(weights)
        counts = [count_for_specified_types * importance for importance in weights]

        weights_multiplier = math.ceil(sum_counts_for_specified_types / sum(counts))
        counts = [count * weights_multiplier for count in counts]

        count_for_topological = size_population - sum_counts_for_specified_types
        count_for_topological = count_for_topological if count_for_topological > 0 else 1
        counts += [count_for_topological]

        chromosome_keys = ['heft_end', 'heft_between', '12.5%', '25%', '75%', '87.5%', 'randomized']
        chromosome_types = self._rand.sample(chromosome_keys, k=size_population, counts=counts)

        chromosomes = self.map(mapper, chromosome_types)
        return [self._toolbox.Individual(chromosome) for chromosome in chromosomes]



from operator import itemgetter
from typing import Optional

import numpy as np
from toposort import CircularDependencyError

from sampo.scheduler.base import SchedulerType
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.utilities.priority import extract_priority_groups_from_nodes
from sampo.schemas.graph import GraphNode
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator


class TopologicalScheduler(GenericScheduler):
    """
    Scheduler, that represent 'WorkGraph' in topological order.
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.Topological,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=AverageReqResourceOptimizer(),
                         timeline_type=MomentumTimeline,
                         prioritization_f=self._topological_sort,
                         optimize_resources_f=self.get_default_res_opt_function(lambda _: Time(0)),
                         work_estimator=work_estimator)

    def _shuffle(self, nodes: set[str]) -> list[str]:
        """
        Shuffle nodes that are on the same level.

        :param nodes: list of nodes
        :return: list of shuffled indices
        """
        return list(nodes)

    # noinspection PyMethodMayBeStatic
    def _topological_sort(self,
                          head_nodes: list[GraphNode],
                          node_id2parent_ids: dict[str, set[str]],
                          node_id2child_ids: dict[str, set[str]],
                          work_estimator: WorkTimeEstimator) -> list[GraphNode]:

        ordered_nodes = []

        priority_groups = extract_priority_groups_from_nodes(head_nodes)

        id2node = {node.id: node for node in head_nodes}

        for _, priority_group in sorted(priority_groups.items(), key=itemgetter(0)):
            priority_group_set = set(node.id for node in priority_group)
            priority_group_dict = {k.id: node_id2parent_ids[k.id].intersection(priority_group_set)
                                   for k in priority_group}
            tsorted_node_ids: list[str] = [node_id
                                           for level in toposort(priority_group_dict)
                                           for node_id in self._shuffle(level)]

            ordered_nodes.extend([id2node[node] for node in tsorted_node_ids])

        # validate_order(ordered_nodes)

        return ordered_nodes


def toposort(data):
    """\
    Dependencies are expressed as a dictionary whose keys are items
    and whose values are a set of dependent items. Output is a list of
    sets in topological order. The first set consists of items with no
    dependences, each subsequent set consists of items that depend upon
    items in the preceeding sets.
    """

    # Special case empty input.
    if len(data) == 0:
        return

    # Copy the input so as to leave it unmodified.
    # Discard self-dependencies and copy two levels deep.
    data = {item: set(e for e in dep if e != item) for item, dep in data.items()}

    # Find all items that don't depend on anything.
    extra_items_in_deps = {value for values in data.values() for value in values} - set(
        data.keys()
    )
    # The line below does N unions of value sets, which is much slower than the
    # set comprehension above which does 1 union of N value sets. The speedup
    # gain is around 200x on a graph with 190k nodes.
    # extra_items_in_deps = _reduce(set.union, data.values()) - set(data.keys())

    # Add empty dependences where needed.
    data.update({item: set() for item in extra_items_in_deps})
    while True:
        ordered = set(item for item, dep in data.items() if len(dep) == 0)
        if not ordered:
            break
        yield ordered
        data = {
            item: (dep - ordered) for item, dep in data.items() if item not in ordered
        }
    if len(data) != 0:
        raise CircularDependencyError(data)


def validate_order(order: list[GraphNode]):
    seen = set()

    for node in order:
        assert all(parent in seen for parent in node.parents)
        seen.add(node)


class RandomizedTopologicalScheduler(TopologicalScheduler):
    """
    Scheduler, that represent 'WorkGraph' in topological order with random.
    """

    def __init__(self,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                 random_seed: Optional[int] = None):
        super().__init__(work_estimator=work_estimator)
        self._random_state = np.random.RandomState(random_seed)

    def _shuffle(self, nodes: set[str]) -> list[str]:
        """
        Shuffle nodes that are on the same level.

        :param nodes: list of nodes
        :return: list of shuffled indices
        """
        nds = list(nodes)
        indices = np.arange(len(nds))
        self._random_state.shuffle(indices)
        return [nds[ind] for ind in indices]



from copy import deepcopy
from dataclasses import dataclass
from typing import Callable

import numpy as np
import torch

from sampo.scheduler.base import Scheduler
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.multi_agency.block_graph import BlockGraph, BlockNode
from sampo.scheduler.multi_agency.exception import NoSufficientAgents
from sampo.scheduler.selection.neural_net import NeuralNetTrainer
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas import WorkerProductivityMode
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.time import Time


class Agent:
    """
    Agent entity representation in the multi-agent model
    Agent have 2 actions: give offer and accept offer
    """

    def __init__(self, name: str, scheduler: Scheduler, contractors: list[Contractor]):
        self.name = name
        self._timeline = None
        self._scheduler = scheduler
        self._contractors = contractors
        self._last_task_executed = Time(0)
        self._downtime = Time(0)

    def offer(self, wg: WorkGraph, parent_time: Time) -> tuple[Time, Time, Schedule, Timeline]:
        """
        Computes the offer from agent to manager. Handles all works from given wg.

        :param wg: the given block of tasks
        :param parent_time: max end time of parent blocks
        :return: offered start time, end time, resulting schedule and timeline before offering

        To apply returned offer, use `Agent#confirm`.
        """
        schedule, start_time, timeline, _ = \
            self._scheduler.schedule_with_cache(wg, self._contractors,
                                                assigned_parent_time=parent_time, timeline=deepcopy(self._timeline))[0]
        return start_time, schedule.execution_time, schedule, timeline

    def confirm(self, timeline: Timeline, start: Time, end: Time):
        """
        Applies the given offer.

        :param timeline: timeline returned from corresponding `Agent#offer`
        :param start: global start time of confirmed block
        :param end: global end time of confirmed block
        """
        self._timeline = timeline
        self.update_stat(start)
        # update last task statistic
        self._last_task_executed = end

    def update_stat(self, start: Time):
        """
        Count last iteration downtime.
        If given start time is lower, than the last executed task, then this downtime are already in self_downtime

        :param start: global start time of confirmed block
        """
        self._downtime += max(Time(0), start - self._last_task_executed)

    def __str__(self) -> str:
        return f'Agent(name={self.name}, scheduler={self._scheduler}, downtime={self._downtime})'

    def __repr__(self) -> str:
        return str(self)

    @property
    def downtime(self) -> Time:
        return self._downtime

    @property
    def contractors(self) -> list[Contractor]:
        return self._contractors

    @property
    def end_time(self) -> Time:
        return self._last_task_executed

    @property
    def scheduler(self) -> Scheduler:
        return self._scheduler


@dataclass
class ScheduledBlock:
    """
    An object represents a scheduled graph block(group of works).

    Contains all data used in scheduling, the agent and resulting information.
    """
    wg: WorkGraph
    schedule: Schedule
    agent: Agent
    start_time: Time
    end_time: Time

    @property
    def id(self) -> str:
        return self.wg.start.id

    @property
    def duration(self) -> Time:
        return self.end_time - self.start_time

    def __str__(self) -> str:
        return f'ScheduledBlock(start_time={self.start_time}, end_time={self.end_time}, agent={self.agent})'

    def __repr__(self) -> str:
        return str(self)


class Manager:
    """
    Manager entity representation in the multi-agent model
    Manager interact with agents

    :param agents: list of agents that has own scheduling algorithm and set of contractors
    """
    def __init__(self, agents: list[Agent]):
        if len(agents) == 0:
            raise NoSufficientAgents('Manager can not work with empty list of agents')
        self._agents = agents

    # TODO Upgrade to supply the best parallelism
    def manage_blocks(self, bg: BlockGraph, logger: Callable[[str], None] = None) -> dict[str, ScheduledBlock]:
        """
        Runs the multi-agent system based on auction on given BlockGraph.
        
        :param bg: 
        :param logger:
        :return: an index of resulting `ScheduledBlock`s built by ids of corresponding `WorkGraph`s
        """
        id2sblock = {}
        for i, block in enumerate(bg.toposort()):
            max_parent_time = max((id2sblock[parent.id].end_time for parent in block.blocks_from), default=Time(0)) + 1
            start_time, end_time, agent_schedule, agent \
                = self.run_auction_with_obstructions(block.wg, max_parent_time, block.obstruction)

            assert start_time >= max_parent_time, f'Scheduler {agent._scheduler} does not handle parent_time!'

            if logger and not block.is_service():
                logger(f'{agent._scheduler}')
            sblock = ScheduledBlock(wg=block.wg, agent=agent, schedule=agent_schedule,
                                    start_time=start_time,
                                    end_time=end_time)
            id2sblock[sblock.id] = sblock

        return id2sblock

    def run_auction_with_obstructions(self, wg: WorkGraph, parent_time: Time = Time(0),
                                      obstruction: Obstruction | None = None):
        if obstruction:
            obstruction.generate(wg)
        return self.run_auction(wg, parent_time)

    def run_auction(self, wg: WorkGraph, parent_time: Time = Time(0)) -> tuple[Time, Time, Schedule, Agent]:
        """
        Runs the auction on the given `WorkGraph`.

        :param wg: target `WorkGraph`
        :param parent_time: max parent time of given block
        :return: best start time, end time and the agent that is able to support this working time
        """
        best_start_time = 0
        best_end_time = Time.inf()
        best_schedule = None
        best_timeline = None
        best_agent = None

        offers = [(agent, agent.offer(wg, parent_time)) for agent in self._agents]

        for offered_agent, (offered_start_time, offered_end_time, offered_schedule, offered_timeline) in offers:
            if offered_end_time < best_end_time:
                best_start_time = offered_start_time
                best_end_time = offered_end_time
                best_schedule = offered_schedule
                best_timeline = offered_timeline
                best_agent = offered_agent
        # To simulate stochastic behaviour
        best_start_time, best_end_time, best_schedule, best_timeline = best_agent.offer(wg, parent_time)
        best_agent.confirm(best_timeline, best_start_time, best_end_time)
        for agent in self._agents:
            if agent.name != best_agent.name:
                agent.update_stat(best_start_time)

        # print(best_agent.name)
        return best_start_time, best_end_time, best_schedule, best_agent


class StochasticManager(Manager):
    """
    Manager entity representation in the multi-agent model
    Manager interact with agents
    """

    def __init__(self, agents: list[Agent]):
        super().__init__(agents)
        self._confidence = {agent.name: 1 for agent in agents}

    def run_auction(self, wg: WorkGraph, parent_time: Time = Time(0)) -> tuple[Time, Time, Schedule, Agent]:
        """
        Runs the auction on the given `WorkGraph`.

        :param wg: target `WorkGraph`
        :param parent_time: max parent time of given block
        :return: best start time, end time and the agent that is able to support this working time
        """
        best_end_time = Time.inf()
        best_agent = None

        def get_offer(agent: Agent):
            agent._scheduler.work_estimator.set_productivity_mode(WorkerProductivityMode.Static)
            return agent, agent.offer(wg, parent_time)

        offers = [get_offer(agent) for agent in self._agents]

        for offered_agent, (offered_start_time, offered_end_time, _, _) in offers:
            offered_end_time = offered_start_time + (offered_end_time - offered_start_time) * self._confidence[offered_agent.name]
            if offered_end_time < best_end_time:
                best_end_time = offered_end_time
                best_agent = offered_agent

        best_agent._scheduler.work_estimator.set_productivity_mode(WorkerProductivityMode.Stochastic)

        old_best_end_time = best_end_time
        best_start_time, best_end_time, best_schedule, best_timeline = best_agent.offer(wg, parent_time)
        modified_end_time = best_start_time + (best_end_time - best_start_time) * self._confidence[
            best_agent.name]
        best_agent.confirm(best_timeline, best_start_time, best_end_time)

        if modified_end_time > old_best_end_time:
            # agent supplied worse time than predicted, lower its confidence
            self._confidence[best_agent.name] += 0.1

        for agent in self._agents:
            if agent.name != best_agent.name:
                agent.update_stat(best_start_time)

        return best_start_time, best_end_time, best_schedule, best_agent


class NeuralManager:
    """
    Manager entity representation in the multi-agent model, that uses neural networks
    Neural manager interact with agents
    Neural manager uses neural network as the method of the most suitable agent for each work graph

    :param agents: list of agents that has own scheduling algorithm and set of contractors
    :param algo_trainer: work with a neural network that predicts the most suitable scheduling algorithm
    :param contractor_trainer: work with a neural network that predicts a set of resources
    that the most suitable contractor should have
    :param algorithms: list of unique scheduling algorithms that agents have
    :param blocks: list of blocks of input Block Graph in topological order
    :param encoding_blocks: list of graph block embeddings
    """

    def __init__(self, agents: list[Agent],
                 algo_trainer: NeuralNetTrainer,
                 contractor_trainer: NeuralNetTrainer,
                 algorithms: list[GenericScheduler],
                 blocks: list[BlockNode],
                 encoding_blocks: list[torch.Tensor]):
        if len(agents) == 0:
            raise NoSufficientAgents('Manager can not work with empty list of agents')
        self._agents = agents
        self.algo_trainer = algo_trainer
        self.contractor_trainer = contractor_trainer
        self.algorithms = algorithms
        self.blocks = blocks
        self.encoding_blocks = encoding_blocks

    # TODO Upgrade to supply the best parallelism
    def manage_blocks(self, logger: Callable[[str], None] = None) -> dict[str, ScheduledBlock]:
        """
        Runs the multi-agent system based on auction on given BlockGraph.

        :param logger:
        :return: an index of resulting `ScheduledBlock`s built by ids of corresponding `WorkGraph`s
        """
        id2sblock = {}
        for i, block in enumerate(self.blocks):
            max_parent_time = max((id2sblock[parent.id].end_time for parent in block.blocks_from), default=Time(0)) + 1
            start_time, end_time, agent_schedule, agent \
                = self.run_auction_with_obstructions(block.wg, i, max_parent_time, block.obstruction)

            assert start_time >= max_parent_time, f'Scheduler {agent._scheduler} does not handle parent_time!'

            if logger and not block.is_service():
                logger(f'{agent._scheduler}')
            sblock = ScheduledBlock(wg=block.wg, agent=agent, schedule=agent_schedule,
                                    start_time=start_time,
                                    end_time=end_time)
            id2sblock[sblock.id] = sblock

        return id2sblock

    def run_auction_with_obstructions(self, wg: WorkGraph,
                                      index: int,
                                      parent_time: Time = Time(0),
                                      obstruction: Obstruction | None = None):
        if obstruction:
            obstruction.generate(wg)
        return self.run_auction(wg, index, parent_time)

    def run_auction(self, wg: WorkGraph, index: int, parent_time: Time = Time(0)) -> tuple[Time, Time, Schedule, Agent]:
        """
        Runs the auction on the given `WorkGraph`.

        :param index: index of agent from the list of agents
        :param wg: target `WorkGraph`
        :param parent_time: max parent time of given block
        :return: best start time, end time and the agent that is able to support this working time
        """

        wg_encoding = [self.encoding_blocks[index]]
        predicted = self.algo_trainer.predict(wg_encoding)
        predict_proba = self.algo_trainer.predict_proba(wg_encoding)
        best_algo = type(self.algorithms[int(predicted)])
        best_contractor = np.asarray(self.contractor_trainer.predict([self.encoding_blocks[index]]))[0]

        time_algo_agents = []
        not_time_algo_agents = []
        time_not_algo_agents = []
        not_time_not_algo_agents = []

        for agent in self._agents:
            if parent_time > agent.end_time and isinstance(agent.scheduler, best_algo):
                time_algo_agents.append(agent)
            elif not isinstance(agent.scheduler, best_algo) and parent_time > agent.end_time:
                time_not_algo_agents.append(agent)
            elif parent_time <= agent.end_time and isinstance(agent.scheduler, best_algo):
                not_time_algo_agents.append(agent)
            else:
                not_time_not_algo_agents.append(agent)

        def auction(agents: list[Agent]) -> Agent:
            less_mse = 10**9
            best_agent = None

            for agent in agents:
                resources = []
                for worker in agent.contractors[0].workers.values():
                    resources.append(worker.count)
                resources = np.asarray(resources)
                mse = sum((resources - best_contractor) ** 2)
                if less_mse > mse:
                    less_mse = mse
                    best_agent = agent

            return best_agent

        if len(time_algo_agents) > 0:
            best_agent = auction(time_algo_agents)
        elif len(time_not_algo_agents) > 0:
            best_agent = auction(time_not_algo_agents)
        elif len(not_time_algo_agents) > 0:
            best_agent = auction(not_time_algo_agents)
        else:
            best_agent = auction(not_time_not_algo_agents)

        best_start_time, best_end_time, best_schedule, best_timeline = best_agent.offer(wg, parent_time)

        best_agent.confirm(best_timeline, best_start_time, best_end_time)
        for agent in self._agents:
            if agent.name != best_agent.name:
                agent.update_stat(best_start_time)

        print(best_agent.name)
        return best_start_time, best_end_time, best_schedule, best_agent



from copy import deepcopy

from sampo.scheduler.base import Scheduler
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time


class AverageBinarySearchResourceOptimizingScheduler:
    """
    The scheduler optimizes resources to deadline.
    Scheduler uses binary search over the resources to optimize them.
    """

    def __init__(self, base_scheduler: Scheduler):
        self._base_scheduler = base_scheduler
        self._resource_optimizer = AverageReqResourceOptimizer()
        base_scheduler.resource_optimizer = self._resource_optimizer

    def schedule_with_cache(self, wg: WorkGraph,
                            contractors: list[Contractor],
                            deadline: Time,
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            assigned_parent_time: Time = Time(0),
                            landscape: LandscapeConfiguration = LandscapeConfiguration()) \
            -> tuple[tuple[Schedule, Time, Timeline, list[GraphNode]], ScheduleSpec]:
        def call_scheduler(k: float, inner_spec: ScheduleSpec) \
                -> tuple[tuple[Schedule, Time, Timeline, list[GraphNode]], ScheduleSpec]:
            self._resource_optimizer.k = k
            try:
                return self._base_scheduler.schedule_with_cache(wg, contractors, inner_spec, validate,
                                                                assigned_parent_time,
                                                                landscape=landscape)[0], inner_spec
            except NoSufficientContractorError:
                return (None, Time.inf(), None, None), inner_spec

        def fitness(k: float, inner_spec: ScheduleSpec):
            result = call_scheduler(k, inner_spec)[0][0].execution_time
            # if result > deadline:
            #     result = Time.inf()
            return result

        copied_spec = deepcopy(spec)
        # FIXME Investigate why `spec` given to this method can be saved from previous call and remove this
        copied_spec._work2spec.clear()

        k_min = 1
        k_max = 10000

        best = k_max

        # steps that we should go further while failing to link the work independent
        try_count = 0
        max_try_count = 3

        result_min_resources = fitness(k_max, copied_spec)
        if result_min_resources < deadline:
            # we can keep the deadline if pass minimum resources,
            # so let's go preventing the works going in parallel
            for node in wg.nodes:
                copied_spec.get_work_spec(node.id).is_independent = True
                new_time = fitness(k_max, copied_spec)
                if new_time > deadline:
                    # if breaking deadline, revert the mutation
                    copied_spec.get_work_spec(node.id).is_independent = False
                    try_count += 1
                    if try_count == max_try_count:
                        # if trying to link independent too many times, break the process
                        break
                else:
                    try_count = 0
            return call_scheduler(best, copied_spec)
        else:
            # we can't keep the deadline if pass minimum resources,
            # but we can pass more

            best_left_fit = Time.inf()
            best_right_fit = Time.inf()
            best_left_m = k_max
            best_right_m = k_min

            while k_max - k_min > 0.05:
                m = (k_min + k_max) / 2
                time_m = fitness(m, copied_spec)
                if time_m > deadline:
                    if abs(time_m.value - deadline.value) <= best_right_fit:
                        best_right_fit = abs(time_m.value - deadline.value)
                        best_right_m = m
                    k_max = m
                else:
                    if abs(time_m.value - deadline.value) <= best_left_fit:
                        best_left_fit = abs(time_m.value - deadline.value)
                        best_left_m = m
                    k_min = m

            if best_left_fit < Time.inf():
                return call_scheduler(best_left_m, copied_spec)
            return call_scheduler(best_right_m, copied_spec)



from datetime import datetime
from functools import partial

import numpy as np
import pandas as pd
from pandas import DataFrame

from sampo.api.const import STAGE_SEP
from sampo.utilities.datetime_util import add_time_delta


def offset_schedule(schedule: DataFrame, offset: datetime | str) -> DataFrame:
    """
    Returns full schedule object with `start` and `finish` columns pushed by date in `offset` argument.
    :param schedule: the schedule itself
    :param offset: Start of schedule, to add as an offset.
    :return: Shifted schedule DataFrame.
    """
    r = schedule.loc[:, :]
    r['start_offset'] = r['start'].apply(partial(add_time_delta, offset))
    r['finish_offset'] = r['finish'].apply(partial(add_time_delta, offset))
    r = r.rename({'start': 'start_', 'finish': 'finish_',
                  'start_offset': 'start', 'finish_offset': 'finish'}, axis=1) \
        .drop(['start_', 'finish_'], axis=1)
    return r


def fix_split_tasks(baps_schedule_df: pd.DataFrame) -> pd.DataFrame:
    """
    Process and merge information for all tasks, which were separated on the several stages during split

    :param baps_schedule_df: pd.DataFrame: schedule with info for tasks separated on stages
    :return: pd.DataFrame: schedule with merged info for all real tasks
    """
    df_len = baps_schedule_df.shape[0]
    baps_schedule_df.index = range(df_len)

    df = pd.DataFrame(columns=baps_schedule_df.columns)
    unique_ids = {x.split(STAGE_SEP)[0] for x in baps_schedule_df.task_id}

    for task_id in unique_ids:
        task_stages_df = baps_schedule_df.loc[
            baps_schedule_df.task_id.str.startswith(f'{task_id}{STAGE_SEP}')
            | (baps_schedule_df.task_id == task_id)
            ]
        task_series = merge_split_stages(task_stages_df.reset_index(drop=True))
        df.loc[df.shape[0]] = task_series  # append

    df = df.sort_values(by=['start', 'task_name'])
    df['idx'] = range(len(df))

    return df


def merge_split_stages(task_df: pd.DataFrame) -> pd.Series:
    """
    Merge split stages of the same real task into one

    :param task_df: pd.DataFrame: one real task's stages dataframe, sorted by start time
    :return: pd.Series with the full information about the task
    """

    def get_stage_num(name: str):
        split_name = name.split(STAGE_SEP)
        return int(split_name[-1]) if len(split_name) > 1 else -1

    if len(task_df) > 1:
        df = task_df.copy()
        df['stage_num'] = df['task_name_mapped'].apply(get_stage_num)
        df = df.sort_values(by='stage_num')
        df = df.reset_index(drop=True)

        df = df.iloc[-1:].reset_index(drop=True)
        for column in ['task_id', 'task_name', 'task_name_mapped']:
            df.loc[0, column] = df.loc[0, column].split(STAGE_SEP)[0]  # fix task id and name

        # sum up volumes through all stages
        df.loc[0, 'volume'] = sum(task_df.loc[:, 'volume'])
        df.loc[0, 'workers'] = task_df.loc[0, 'workers']

        # fix task's start time and duration
        df.loc[0, 'start'] = task_df.loc[0, 'start']
        df.loc[0, 'finish'] = task_df.loc[len(task_df) - 1, 'finish']
        if isinstance(df.loc[0, 'start'], np.int64) or isinstance(df.loc[0, 'start'], np.int32):
            df.loc[0, 'duration'] = df.loc[0, 'finish'] - df.loc[0, 'start'] + 1
        else:
            df.loc[0, 'duration'] = (df.loc[0, 'finish'] - df.loc[0, 'start']).days + 1
    else:
        df = task_df.copy()

    return df.loc[0, :]



from sampo.scheduler.selection.neural_net import NeuralNet, NeuralNetTrainer, NeuralNetType
from sampo.scheduler.selection.validation import cross_val_score
from sampo.scheduler.selection.metrics import encode_graph



from collections import deque

from sortedcontainers import SortedList

from sampo.schemas.requirements import ZoneReq
from sampo.schemas.time import Time
from sampo.schemas.types import EventType, ScheduleEvent
from sampo.schemas.zones import ZoneConfiguration, Zone, ZoneTransition
from sampo.utilities.collections_util import build_index


class ZoneTimeline:

    def __init__(self, config: ZoneConfiguration):
        def event_cmp(event: ScheduleEvent | Time | tuple[Time, int, int]) -> tuple[Time, int, int]:
            if isinstance(event, ScheduleEvent):
                if event.event_type is EventType.INITIAL:
                    return Time(-1), -1, event.event_type.priority

                return event.time, event.seq_id, event.event_type.priority

            if isinstance(event, Time):
                # instances of Time must be greater than almost all ScheduleEvents with same time point
                return event, Time.inf().value, 2

            if isinstance(event, tuple):
                return event

            raise ValueError(f'Incorrect type of value: {type(event)}')

        self._timeline = {zone: SortedList([ScheduleEvent(-1, EventType.INITIAL, Time(0), None, status)],
                                           key=event_cmp)
                          for zone, status in config.start_statuses.items()}
        self._config = config

    def find_min_start_time(self, zones: list[ZoneReq], parent_time: Time, exec_time: Time) -> Time:
        # here we look for the earliest time slot that can satisfy all the zones

        start = parent_time
        scheduled_wreqs: list[ZoneReq] = []

        type2status: dict[str, int] = build_index(zones, lambda w: w.kind, lambda w: w.required_status)

        queue = deque(zones)

        i = 0
        while len(queue) > 0:
            # This should be uncommented when there are problems with performance

            # if i > 0 and i % 50 == 0:
            #     print(f'Warning! Probably cycle in looking for time slot for all reqs: {i} iteration')
            #     print(f'Current queue size: {len(queue)}')
            i += 1

            wreq = queue.popleft()
            state = self._timeline[wreq.kind]
            # we look for the earliest time slot starting from 'start' time moment
            # if we have found a time slot for the previous task,
            # we should start to find for the earliest time slot of other task since this new time
            found_start = self._find_earliest_time_slot(state, start, exec_time, type2status[wreq.kind])

            assert found_start >= start

            if len(scheduled_wreqs) == 0 or start == found_start:
                # we schedule the first worker's specialization or the next spec has the same start time
                # as the all previous ones
                scheduled_wreqs.append(wreq)
                start = max(start, found_start)
            else:
                # The current worker specialization can be started only later than
                # the previously found start time.
                # In this case we need to add back all previously scheduled wreq-s into the queue
                # to be scheduled again with the new start time (e.g. found start).
                # This process should reach its termination at least at the very end of this contractor's schedule.
                queue.extend(scheduled_wreqs)
                scheduled_wreqs.clear()
                scheduled_wreqs.append(wreq)
                start = max(start, found_start)

        # This should be uncommented when there are problems with zone scheduling correctness
        # for w in zones:
        #     self._validate(start, exec_time, self._timeline[w.kind], w.required_status)

        return start

    def _match_status(self, target: int, match: int) -> bool:
        return self._config.statuses.match_status(target, match)

    def _validate(self, start_time: Time, exec_time: Time, state: SortedList[ScheduleEvent], required_status: int):
        # === THE INNER VALIDATION ===

        start_idx = state.bisect_right(start_time)
        end_idx = state.bisect_right(start_time + exec_time)
        start_status = state[start_idx - 1].available_workers_count

        # updating all events in between the start and the end of our current task
        for event in state[start_idx: end_idx]:
            assert self._config.statuses.match_status(event.available_workers_count, required_status)

        assert state[start_idx - 1].event_type == EventType.END \
               or (state[start_idx - 1].event_type == EventType.START
                   and self._config.statuses.match_status(start_status, required_status)) \
               or state[start_idx - 1].event_type == EventType.INITIAL, \
            f'{state[start_idx - 1].time} {state[start_idx - 1].event_type} {required_status} {start_status}'

        # === END OF INNER VALIDATION ===

    def _find_earliest_time_slot(self,
                                 state: SortedList[ScheduleEvent],
                                 parent_time: Time,
                                 exec_time: Time,
                                 required_status: int) -> Time:
        """
        Searches for the earliest time starting from start_time, when a time slot
        of exec_time is available, when required_worker_count of resources is available

        :param state: stores Timeline for the certain resource
        :param parent_time: the minimum start time starting from the end of the parent task
        :param exec_time: execution time of work
        :param required_status: requirements status of zone
        :return: the earliest start time
        """
        current_start_time = parent_time
        current_start_idx = state.bisect_right(current_start_time) - 1

        # the condition means we have reached the end of schedule for this contractor subject to specialization (wreq)
        # as long as we assured that this contractor has enough capacity at all to handle the task
        # we can stop and put the task at the very end
        i = 0
        while len(state[current_start_idx:]) > 0:
            if i > 0 and i % 50 == 0:
                print(f'Warning! Probably cycle in looking for earliest time slot: {i} iteration')
                print(f'Current start time: {current_start_time}, current start idx: {current_start_idx}')
            i += 1

            current_start_status = state[current_start_idx].available_workers_count
            end_idx = state.bisect_right(current_start_time + exec_time)

            if not self._match_status(current_start_status, required_status):
                if current_start_idx == len(state) - 1:
                    current_start_time += max(0, self._config.time_costs[current_start_status, required_status]
                                              - (current_start_time - state[-1].time))
                    break  # if we are in the very end, break

                # if we are inside the interval with wrong status
                # we should go right and search the better begin
                if state[current_start_idx].event_type == EventType.START \
                        and not self._match_status(current_start_status, required_status):
                    current_start_idx += 1
                    current_start_time = state[current_start_idx].time
                    continue

                # here we are outside the all intervals or inside the interval with right status
                # if we are outside intervals, we can be in right or wrong status, so let's check it
                # else we are inside the interval with right status so let

                if self._is_inside_interval(state, current_start_idx):
                    if not self._match_status(current_start_status, required_status):
                        # we are inside the interval with wrong status, so we can't change it, go next
                        current_start_idx += 1
                        current_start_time = state[current_start_idx].time
                        continue
                else:
                    # we are outside the interval, should change status
                    # check that we can do it
                    current_start_time += self._config.time_costs[current_start_status, required_status]
                    end_idx = state.bisect_right(current_start_time + exec_time)

            # here we are guaranteed that current_start_time is in right status
            # so go right and check matching statuses
            # this step performed like in MomentumTimeline
            not_compatible_status_found = False
            for idx in range(end_idx - 1, current_start_idx - 1, -1):
                if not self._match_status(state[idx].available_workers_count, required_status):
                    # we're trying to find a new slot that would start with
                    # either the last index passing the quantity check
                    # or the index after the execution interval
                    # we need max here to process a corner case when the problem arises
                    # on current_start_idx - 1
                    # without max it would get into infinite cycle
                    current_start_idx = max(idx, current_start_idx) + 1
                    not_compatible_status_found = True
                    break

            if not not_compatible_status_found:
                break

            if current_start_idx >= len(state):
                # This should be uncommented when there are problems with zone scheduling correctness

                # cur_cpkt = state[-1]
                # if cur_cpkt.time == current_start_time and not self._match_status(cur_cpkt.available_workers_count,
                #                                                                   required_status):
                #     # print('Problem!')
                #     current_start_time = max(parent_time, state[-1].time + 1)
                current_start_time = max(parent_time, state[-1].time + 1)
                break

            current_start_time = state[current_start_idx].time

        # This should be uncommented when there are problems with zone scheduling correctness
        self._validate(current_start_time, exec_time, state, required_status)

        return current_start_time

    # noinspection PyMethodMayBeStatic
    def _is_inside_interval(self, state: SortedList[ScheduleEvent], idx: int) -> bool:
        # TODO Make better algorithms to check that we can change status in point `start_time - change_cost`
        starts_count = 0
        ends_count = 0
        # checking from end because it's more realistic to call this function at the almost end of timeline
        for cpkt in state[idx:]:
            if cpkt.event_type == EventType.START:
                starts_count += 1
            else:
                ends_count += 1
        # we are inside the interval with incompatible status, break
        return starts_count == ends_count

    def can_schedule_at_the_moment(self, zones: list[ZoneReq], start_time: Time, exec_time: Time):
        for zone in zones:
            state = self._timeline[zone.kind]

            start_idx = state.bisect_right(start_time)
            end_idx = state.bisect_right(start_time + exec_time)
            start_status = state[start_idx - 1].available_workers_count

            if not self._match_status(start_status, zone.required_status):
                # starting status don't match, trying to change status
                change_cost = self._config.time_costs[start_status, zone.required_status]
                new_start_idx = state.bisect_right(start_time - change_cost)
                if new_start_idx != start_idx:
                    # we have incompatible status inside our interval, break
                    return False

                if not self._is_inside_interval(state, start_idx):
                    return False

            # checking all events in between the start and the end of our current task
            for event in state[start_idx: end_idx]:
                if not self._match_status(event.available_workers_count, zone.required_status):
                    return False

        return True

    def update_timeline(self, index: int, zones: list[Zone], start_time: Time, exec_time: Time) -> list[ZoneTransition]:
        sworks = []

        for zone in zones:
            state = self._timeline[zone.name]
            start_idx = state.bisect_right(start_time)
            start_status = state[start_idx - 1].available_workers_count

            # This should be uncommented when there are problems with zone scheduling correctness
            self._validate(start_time, exec_time, state, zone.status)

            change_cost = self._config.time_costs[start_status, zone.status] \
                if not self._config.statuses.match_status(start_status, zone.status) \
                else 0

            state.add(ScheduleEvent(index, EventType.START, start_time - change_cost, None, zone.status))
            state.add(ScheduleEvent(index, EventType.END, start_time - change_cost + exec_time, None, zone.status))

            if start_status != zone.status and zone.status != 0:
                # if we need to change status, record it
                sworks.append(ZoneTransition(name=zone.name,
                                             from_status=start_status,
                                             to_status=zone.status,
                                             start_time=start_time - change_cost,
                                             end_time=start_time))
        return sworks

    def append_statuses(self, zones: list[Zone]) -> tuple[Time, list[ZoneTransition]]:
        global_finish_time = Time(0)
        sworks = []

        for zone in zones:
            state = self._timeline[zone.name]
            latest_time = state[-1].time + 1
            latest_status = state[-1].available_workers_time
            change_cost = self._config.time_costs[latest_status, zone.status] \
                if not self._config.statuses.match_status(latest_status, zone.status) \
                else 0
            finish_time = latest_time + change_cost

            state.add(ScheduleEvent(Time.inf().value, EventType.START, latest_time, None, zone.status))
            state.add(ScheduleEvent(Time.inf().value, EventType.END, finish_time, None, zone.status))

            if latest_status != zone.status and zone.status != 0:
                # if we need to change status, record it
                sworks.append(ZoneTransition(name=zone.name,
                                             from_status=latest_status,
                                             to_status=zone.status,
                                             start_time=latest_time - change_cost,
                                             end_time=latest_time))

            global_finish_time = max(global_finish_time, finish_time)

        return global_finish_time, sworks

    def finish_statuses(self) -> tuple[Time, list[ZoneTransition]]:
        return self.append_statuses([Zone(*v) for v in self._config.end_statuses])



from dataclasses import dataclass

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.requirements import WorkerReq, EquipmentReq, MaterialReq, ConstructionObjectReq, ZoneReq
from sampo.schemas.resources import Material
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.utilities.serializers import custom_serializer


@dataclass
class WorkUnit(AutoJSONSerializable['WorkUnit'], Identifiable):
    """
    Class that describe vertex in graph (one work/task)
    """
    def __init__(self,
                 id: str,
                 name: str,
                 worker_reqs: list[WorkerReq] = None,
                 equipment_reqs: list[EquipmentReq] = None,
                 material_reqs: list[MaterialReq] = None,
                 object_reqs: list[ConstructionObjectReq] = None,
                 zone_reqs: list[ZoneReq] = None,
                 description: str = '',
                 group: str = 'main project',
                 priority: int = 1,
                 is_service_unit: bool = False,
                 volume: float = 0,
                 volume_type: str = 'unit',
                 display_name: str = "",
                 workground_size: int = 100):
        """
        :param worker_reqs: list of required professions (i.e. workers)
        :param equipment_reqs: list of required equipment
        :param material_reqs: list of required materials (e.g. logs, stones, gravel etc.)
        :param object_reqs: list of required objects (e.g. electricity, pipelines, roads)
        :param zone_reqs: list of required zone statuses (e.g. opened/closed doors, attached equipment, etc.)
        :param description: the description. It is useful, for example, to show it on visualization
        :param group: union block of works
        :param is_service_unit: service units are additional vertexes
        :param volume: scope of work
        :param volume_type: unit of scope of work
        :param display_name: name of work
        """
        super(WorkUnit, self).__init__(id, name)
        if material_reqs is None:
            material_reqs = []
        if object_reqs is None:
            object_reqs = []
        if equipment_reqs is None:
            equipment_reqs = []
        if worker_reqs is None:
            worker_reqs = []
        if zone_reqs is None:
            zone_reqs = []
        self.worker_reqs = worker_reqs
        self.equipment_reqs = equipment_reqs
        self.object_reqs = object_reqs
        self.material_reqs = material_reqs
        self.zone_reqs = zone_reqs
        self.description = description
        self.group = group
        self.is_service_unit = is_service_unit
        self.volume = float(volume)
        self.volume_type = volume_type
        self.display_name = display_name if display_name else name
        self.priority = priority

    def __del__(self):
        for attr in self.__dict__.values():
            del attr

    def need_materials(self) -> list[Material]:
        return [req.material() for req in self.material_reqs]

    @custom_serializer('worker_reqs')
    @custom_serializer('zone_reqs')
    @custom_serializer('material_reqs')
    def serialize_serializable_list(self, value) -> list:
        """
        Return serialized list of values.
        Values should be serializable.

        :param value: list of values
        :return: list of serialized values
        """
        return [t._serialize() for t in value]

    @classmethod
    @custom_serializer('material_reqs', deserializer=True)
    def material_reqs_deserializer(cls, value) -> list[MaterialReq]:
        """
        Get list of material requirements

        :param value: serialized list of material requirements
        :return: list of material requirements
        """
        return [MaterialReq._deserialize(wr) for wr in value]

    @classmethod
    @custom_serializer('worker_reqs', deserializer=True)
    def worker_reqs_deserializer(cls, value) -> list[WorkerReq]:
        """
        Get list of worker requirements

        :param value: serialized list of work requirements
        :return: list of worker requirements
        """
        return [WorkerReq._deserialize(wr) for wr in value]

    @classmethod
    @custom_serializer('zone_reqs', deserializer=True)
    def zone_reqs_deserializer(cls, value) -> list[ZoneReq]:
        """
        Get list of worker requirements

        :param value: serialized list of work requirements
        :return: list of worker requirements
        """
        return [ZoneReq._deserialize(wr) for wr in value]

    @classmethod
    @custom_serializer('material_reqs', deserializer=True)
    def material_reqs_deserializer(cls, value) -> list[MaterialReq]:
        """
        Get list of material requirements

        :param value: serialized list of material requirements
        :return: list of material requirements
        """
        return [MaterialReq._deserialize(wr) for wr in value]

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        return self._serialize()

    def __setstate__(self, state):
        new_work_unit = self._deserialize(state)
        self.worker_reqs = new_work_unit.worker_reqs
        self.equipment_reqs = new_work_unit.equipment_reqs
        self.object_reqs = new_work_unit.object_reqs
        self.material_reqs = new_work_unit.material_reqs
        self.zone_reqs = new_work_unit.zone_reqs
        self.id = new_work_unit.id
        self.name = new_work_unit.name
        self.is_service_unit = new_work_unit.is_service_unit
        self.volume = new_work_unit.volume
        self.volume_type = new_work_unit.volume_type
        self.group = new_work_unit.group
        self.display_name = new_work_unit.display_name
        self.priority = new_work_unit.priority



class WorkGraphBuildingException(Exception):
    """
    Raised when work graph can't be built
    """
    def __init__(self, message: str):
        super().__init__(message)


class InputDataException(Exception):
    """
    Raise when there isn't info about tasks' connection
    """
    def __init__(self, reason: str):
        super().__init__("The reason of Input Data exception is " + reason)



import copy

import numpy as np

from sampo.api.genetic_api import ChromosomeType, ScheduleGenerationScheme
from sampo.base import SAMPO
from sampo.scheduler.base import Scheduler
from sampo.scheduler.timeline import JustInTimeTimeline, MomentumTimeline
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.general_timeline import GeneralTimeline
from sampo.scheduler.utils import WorkerContractorPool
from sampo.scheduler.utils.time_computaion import calculate_working_time_cascade
from sampo.schemas import ZoneReq
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduledWork, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.linked_list import LinkedList


def convert_schedule_to_chromosome(work_id2index: dict[str, int],
                                   worker_name2index: dict[str, int],
                                   contractor2index: dict[str, int],
                                   contractor_borders: np.ndarray,
                                   schedule: Schedule,
                                   spec: ScheduleSpec,
                                   landscape: LandscapeConfiguration,
                                   order: list[GraphNode] | None = None) -> ChromosomeType:
    """
    Receive a result of scheduling algorithm and transform it to chromosome

    :param work_id2index:
    :param worker_name2index:
    :param contractor2index:
    :param contractor_borders:
    :param schedule:
    :param spec:
    :param landscape:
    :param order: if passed, specify the node order that should appear in the chromosome
    :return:
    """

    # order: list[GraphNode] = order if order is not None else [work for work in schedule.works
    #                                                           if work.id in work_id2index]

    # order works part of chromosome
    order_chromosome: np.ndarray = np.array([work_id2index[work.id] for work in order])

    # convert to convenient form
    schedule = schedule.to_schedule_work_dict

    # resources for works part of chromosome
    # +1 stores contractors line
    resource_chromosome = np.zeros((len(order_chromosome), len(worker_name2index) + 1), dtype=int)

    # zone status changes after node executing
    zone_changes_chromosome = np.zeros((len(order_chromosome), len(landscape.zone_config.start_statuses)), dtype=int)

    for node in order:
        node_id = node.id
        index = work_id2index[node_id]
        for resource in schedule[node_id].workers:
            res_count = resource.count
            res_index = worker_name2index[resource.name]
            res_contractor = resource.contractor_id
            resource_chromosome[index, res_index] = res_count
            resource_chromosome[index, -1] = contractor2index[res_contractor]

    resource_border_chromosome = np.copy(contractor_borders)

    return order_chromosome, resource_chromosome, resource_border_chromosome, spec, zone_changes_chromosome


def convert_chromosome_to_schedule(chromosome: ChromosomeType,
                                   worker_pool: WorkerContractorPool,
                                   index2node: dict[int, GraphNode],
                                   index2contractor: dict[int, Contractor],
                                   index2zone: dict[int, str],
                                   worker_pool_indices: dict[int, dict[int, Worker]],
                                   worker_name2index: dict[str, int],
                                   contractor2index: dict[str, int],
                                   landscape: LandscapeConfiguration,
                                   timeline: Timeline | None = None,
                                   assigned_parent_time: Time = Time(0),
                                   work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                                   sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline, list[GraphNode]]:
    """
    Build schedule from received chromosome
    It can be used in visualization of final solving of genetic algorithm
    """
    match sgs_type:
        case ScheduleGenerationScheme.Parallel:
            converter = parallel_schedule_generation_scheme
        case ScheduleGenerationScheme.Serial:
            converter = serial_schedule_generation_scheme
        case _:
            raise ValueError('Unknown type of schedule generation scheme')
    return converter(chromosome,
                     worker_pool,
                     index2node,
                     index2contractor,
                     index2zone,
                     worker_pool_indices,
                     worker_name2index,
                     contractor2index,
                     landscape,
                     timeline,
                     assigned_parent_time,
                     work_estimator)


def parallel_schedule_generation_scheme(chromosome: ChromosomeType,
                                        worker_pool: WorkerContractorPool,
                                        index2node: dict[int, GraphNode],
                                        index2contractor: dict[int, Contractor],
                                        index2zone: dict[int, str],
                                        worker_pool_indices: dict[int, dict[int, Worker]],
                                        worker_name2index: dict[str, int],
                                        contractor2index: dict[str, int],
                                        landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                        timeline: Timeline | None = None,
                                        assigned_parent_time: Time = Time(0),
                                        work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline, list[GraphNode]]:
    """
    Implementation of Parallel Schedule Generation Scheme
    """
    node2swork: dict[GraphNode, ScheduledWork] = {}

    works_order = chromosome[0]
    works_resources = chromosome[1]
    border = chromosome[2]
    spec = chromosome[3]
    zone_statuses = chromosome[4]
    worker_pool = copy.deepcopy(worker_pool)

    # use 3rd part of chromosome in schedule generator
    for worker_index in worker_pool:
        for contractor_index in worker_pool[worker_index]:
            worker_pool[worker_index][contractor_index].with_count(border[contractor2index[contractor_index],
            worker_name2index[worker_index]])

    if not isinstance(timeline, JustInTimeTimeline):
        timeline = JustInTimeTimeline(worker_pool, landscape)

    order_nodes = []

    # timeline to store starts and ends of all works
    work_timeline = GeneralTimeline()

    def decode(work_index):
        cur_node = index2node[work_index]

        cur_work_spec = spec.get_work_spec(cur_node.id)
        cur_resources = works_resources[work_index, :-1]
        cur_contractor_index = works_resources[work_index, -1]
        cur_contractor = index2contractor[cur_contractor_index]
        cur_worker_team: list[Worker] = [worker_pool_indices[worker_index][cur_contractor_index]
                                         .copy().with_count(worker_count)
                                         for worker_index, worker_count in enumerate(cur_resources)
                                         if worker_count > 0]
        if cur_work_spec.assigned_time is not None:
            cur_exec_time = cur_work_spec.assigned_time
        else:
            cur_exec_time = calculate_working_time_cascade(cur_node, cur_worker_team, work_estimator)
        return cur_node, cur_worker_team, cur_contractor, cur_exec_time, cur_work_spec

    # account the remaining works
    enumerated_works_remaining = LinkedList(iterable=enumerate(
        [(work_index, *decode(work_index)) for work_index in works_order]
    ))

    # declare current checkpoint index
    ckpt_idx = 0
    start_time = assigned_parent_time - 1
    prev_start_time = start_time - 1

    def work_scheduled(args) -> bool:
        idx, (work_idx, node, worker_team, contractor, exec_time, work_spec) = args

        if timeline.can_schedule_at_the_moment(node, worker_team, work_spec, node2swork, start_time, exec_time):
            # apply worker spec
            Scheduler.optimize_resources_using_spec(node.work_unit, worker_team, work_spec)

            st = start_time
            if idx == 0:  # we are scheduling the work `start of the project`
                st = assigned_parent_time  # this work should always have st = 0, so we just re-assign it

            if idx == len(works_order) - 1:  # we are scheduling the work `end of the project`
                finish_time, finalizing_zones = timeline.zone_timeline.finish_statuses()
                st = max(start_time, finish_time)

            # finish using time spec
            timeline.schedule(node, node2swork, worker_team, contractor, work_spec,
                              st, exec_time, assigned_parent_time, work_estimator)

            if idx == len(works_order) - 1:  # we are scheduling the work `end of the project`
                node2swork[node].zones_pre = finalizing_zones

            work_timeline.update_timeline(st, exec_time, None)
            return True
        return False

    # while there are unprocessed checkpoints
    while len(enumerated_works_remaining) > 0:
        if ckpt_idx < len(work_timeline):
            start_time = work_timeline[ckpt_idx]
            if prev_start_time == start_time:
                ckpt_idx += 1
                continue
            if start_time.is_inf():
                # break because schedule already contains Time.inf(), that is incorrect schedule
                break
            prev_start_time = start_time
        else:
            start_time += 1

        # find all works that can start at start_time moment and remove it if scheduled
        enumerated_works_remaining.remove_if(work_scheduled)
        ckpt_idx = min(ckpt_idx + 1, len(work_timeline))

    return node2swork, assigned_parent_time, timeline, order_nodes


def serial_schedule_generation_scheme(chromosome: ChromosomeType,
                                      worker_pool: WorkerContractorPool,
                                      index2node: dict[int, GraphNode],
                                      index2contractor: dict[int, Contractor],
                                      index2zone: dict[int, str],
                                      worker_pool_indices: dict[int, dict[int, Worker]],
                                      worker_name2index: dict[str, int],
                                      contractor2index: dict[str, int],
                                      landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                      timeline: Timeline | None = None,
                                      assigned_parent_time: Time = Time(0),
                                      work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline, list[GraphNode]]:
    """
    Implementation of Serial Schedule Generation Scheme
    """
    node2swork: dict[GraphNode, ScheduledWork] = {}

    works_order = chromosome[0]
    works_resources = chromosome[1]
    border = chromosome[2]
    spec = chromosome[3]
    zone_statuses = chromosome[4]
    worker_pool = copy.deepcopy(worker_pool)

    # use 3rd part of chromosome in schedule generator
    for worker_index in worker_pool:
        for contractor_index in worker_pool[worker_index]:
            worker_pool[worker_index][contractor_index].with_count(border[contractor2index[contractor_index],
            worker_name2index[worker_index]])

    if not isinstance(timeline, MomentumTimeline):
        timeline = MomentumTimeline(worker_pool, landscape)

    order_nodes = []

    for order_index, work_index in enumerate(works_order):
        node = index2node[work_index]
        order_nodes.append(node)

        work_spec = spec.get_work_spec(node.id)

        resources = works_resources[work_index, :-1]
        contractor_index = works_resources[work_index, -1]
        contractor = index2contractor[contractor_index]
        worker_team: list[Worker] = [worker_pool[wreq.kind][contractor.id]
                                     .copy().with_count(resources[worker_name2index[wreq.kind]])
                                     for wreq in node.work_unit.worker_reqs]

        # apply worker spec
        Scheduler.optimize_resources_using_spec(node.work_unit, worker_team, work_spec)

        st, ft, exec_times = timeline.find_min_start_time_with_additional(node, worker_team, node2swork, work_spec,
                                                                          assigned_parent_time=assigned_parent_time,
                                                                          work_estimator=work_estimator)

        if order_index == 0:  # we are scheduling the work `start of the project`
            st = assigned_parent_time  # this work should always have st = 0, so we just re-assign it

        # finish using time spec
        ft = timeline.schedule(node, node2swork, worker_team, contractor, work_spec,
                               st, work_spec.assigned_time, assigned_parent_time, work_estimator)
        # process zones
        zone_reqs = [ZoneReq(index2zone[i], zone_status) for i, zone_status in enumerate(zone_statuses[work_index])]
        zone_start_time = timeline.zone_timeline.find_min_start_time(zone_reqs, ft, 0)

        # we should deny scheduling
        # if zone status change can be scheduled only in delayed manner
        if zone_start_time != ft:
            node2swork[node].zones_post = timeline.zone_timeline.update_timeline(order_index,
                                                                                 [z.to_zone() for z in zone_reqs],
                                                                                 zone_start_time, 0)

    schedule_start_time = min((swork.start_time for swork in node2swork.values() if
                               len(swork.workers) != 0), default=assigned_parent_time)

    return node2swork, schedule_start_time, timeline, order_nodes



import uuid

from sortedcontainers import SortedList

from sampo.schemas.graph import GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.landscape_graph import LandGraphNode
from sampo.schemas.resources import Material
from sampo.schemas.time import Time
from sampo.schemas.types import ScheduleEvent, EventType


class PlatformTimeline:
    def __init__(self, landscape_config: LandscapeConfiguration):

        def event_cmp(event: ScheduleEvent | Time | tuple[Time, int, int]) -> tuple[Time, int, int]:
            if isinstance(event, ScheduleEvent):
                if event.event_type is EventType.INITIAL:
                    return Time(-1), -1, event.event_type.priority

                return event.time, event.seq_id, event.event_type.priority

            if isinstance(event, Time):
                # instances of Time must be greater than almost all ScheduleEvents with same time point
                return event, Time.inf().value, 2

            if isinstance(event, tuple):
                return event

            raise ValueError(f'Incorrect type of value: {type(event)}')

        self._timeline: dict[str, dict[str, SortedList[ScheduleEvent]]] = {}
        self._task_index = 0
        self._landscape = landscape_config
        for mat_id, mat_dict in landscape_config.get_platforms_resources().items():
            self._timeline[mat_id] = {
                mat[0]: SortedList(iterable=(ScheduleEvent(-1, EventType.INITIAL, Time(0), None, mat[1]),),
                                   key=event_cmp)
                for mat in mat_dict.items()
            }

    def can_schedule_at_the_moment(self, node: GraphNode, start_time: Time,
                                   materials: list[Material]) -> bool:
        if not node.work_unit.need_materials():
            return True

        platform = self._landscape.works2platform[node]
        if not self._check_material_availability_on_platform(platform, materials, start_time):
            return False

        return True

    def _check_material_availability_on_platform(self, platform: LandGraphNode, materials: list[Material],
                                                 start_time: Time) -> bool:
        """
        Check the materials' availability on the `platform` at the `start_time`
        """
        # TODO Add delivery opportunity checking
        platform_state = self._timeline[platform.id]

        for mat in materials:
            start = platform_state[mat.name].bisect_right(start_time)
            finish = len(platform_state[mat.name])

            if finish - start > 1:
                return False

        return True

    def get_material_for_delivery(self, node: GraphNode, materials: list[Material], work_start_time: Time) \
            -> list[Material]:
        """
        Returns `Material`s that should be delivered to the `node`'s platform to start the corresponding work
        """
        request: list[Material] = []
        platform = self._landscape.works2platform[node]
        platform_state = self._timeline[platform.id]
        for need_mat in materials:
            start = platform_state[need_mat.name].bisect_right(work_start_time) - 1
            available_count_material = platform_state[need_mat.name][start].available_workers_count

            if available_count_material < need_mat.count:
                request.append(
                    Material(str(uuid.uuid4()), need_mat.name,
                             self._landscape.works2platform[node].resource_storage_unit.capacity[need_mat.name] -
                             available_count_material)
                )

        return request

    def find_min_material_time_with_additional(self, node: GraphNode, start_time: Time,
                                               materials: list[Material]) -> tuple[Time, list[Material]]:
        platform = self._landscape.works2platform[node]
        platform_state = self._timeline[platform.id]
        if not self._check_material_availability_on_platform(platform, materials, start_time):
            for mat in materials:
                ind = len(platform_state[mat.name]) - 1
                mat_last_time = platform_state[mat.name][ind].time
                start_time = max(mat_last_time, start_time)

        mat_request = self.get_material_for_delivery(node, materials, start_time)
        return start_time, mat_request

    def can_provide_resources(self,
                              node: GraphNode,
                              deadline: Time,
                              materials: list[Material]) -> bool:

        start_time = deadline

        materials_for_delivery = self.get_material_for_delivery(node, materials, start_time)
        platform = self._landscape.works2platform[node]
        # TODO Simplify OR because it checks emptiness of materials for delivery
        if not materials_for_delivery and self._check_material_availability_on_platform(platform, materials, start_time):
            update_timeline_info: list[tuple[str, int, Time]] = [(mat.name, mat.count, start_time) for mat in materials]
            self.update_timeline(platform.id, update_timeline_info)

            return True
        return False

    def update_timeline(self, platform_id: str, update_timeline_info: list[tuple[str, int, Time]]) -> None:
        res_holder_state = self._timeline[platform_id]

        for res_info in update_timeline_info:
            task_index = self._task_index
            self._task_index += 1

            res_name, res_count, start_time = res_info
            res_state = res_holder_state[res_name]
            start_idx = res_state.bisect_right(start_time)

            available_res_count = res_state[start_idx - 1].available_workers_count

            res_state.add(
                ScheduleEvent(task_index, EventType.START, start_time, None, available_res_count - res_count)
            )



from random import Random
from typing import Callable

import sampo.scheduler

from sampo.api.genetic_api import FitnessFunction, ChromosomeType, Individual, ScheduleGenerationScheme
from sampo.backend import ComputationalBackend, T, R
from sampo.schemas import WorkGraph, Contractor, LandscapeConfiguration, WorkTimeEstimator, Schedule, GraphNode, Time
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time_estimator import DefaultWorkEstimator


class DefaultComputationalBackend(ComputationalBackend):

    def map(self, action: Callable[[T], R], values: list[T]) -> list[R]:
        return [action(v) for v in values]

    def cache_scheduler_info(self,
                             wg: WorkGraph,
                             contractors: list[Contractor],
                             landscape: LandscapeConfiguration = LandscapeConfiguration(),
                             spec: ScheduleSpec = ScheduleSpec(),
                             rand: Random | None = None,
                             work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        self._wg = wg
        self._contractors = contractors
        self._landscape = landscape
        self._spec = spec
        self._rand = rand
        self._work_estimator = work_estimator
        self._toolbox = None

    def cache_genetic_info(self,
                           population_size: int = 50,
                           mutate_order: float = 0.1,
                           mutate_resources: float = 0.05,
                           mutate_zones: float = 0.05,
                           deadline: Time | None = None,
                           weights: list[int] | None = None,
                           init_schedules: dict[str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]] = None,
                           assigned_parent_time: Time = Time(0),
                           fitness_weights: tuple[int | float, ...] = None,
                           sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                           only_lft_initialization: bool = False,
                           is_multiobjective: bool = False):
        self._selection_size = population_size
        self._mutate_order = mutate_order
        self._mutate_resources = mutate_resources
        self._mutate_zones = mutate_zones
        self._deadline = deadline
        self._weights = weights
        self._init_schedules = init_schedules
        self._assigned_parent_time = assigned_parent_time
        self._fitness_weights = fitness_weights
        self._sgs_type = sgs_type
        self._only_lft_initialization = only_lft_initialization
        self._is_multiobjective = is_multiobjective
        self._toolbox = None

    def _ensure_toolbox_created(self):
        if self._toolbox is None:
            from sampo.scheduler.genetic.utils import init_chromosomes_f, create_toolbox_using_cached_chromosomes

            if self._init_schedules:
                init_chromosomes = init_chromosomes_f(self._wg, self._contractors, self._init_schedules,
                                                      self._landscape)
            else:
                init_chromosomes = []

            rand = self._rand or Random()
            work_estimator = self._work_estimator or DefaultWorkEstimator()
            assigned_parent_time = self._assigned_parent_time or Time(0)

            self._toolbox = create_toolbox_using_cached_chromosomes(self._wg,
                                                                    self._contractors,
                                                                    self._selection_size,
                                                                    self._mutate_order,
                                                                    self._mutate_resources,
                                                                    self._mutate_zones,
                                                                    init_chromosomes,
                                                                    rand,
                                                                    self._spec,
                                                                    work_estimator,
                                                                    assigned_parent_time,
                                                                    self._fitness_weights,
                                                                    self._landscape,
                                                                    self._sgs_type,
                                                                    self._only_lft_initialization,
                                                                    self._is_multiobjective)

    def compute_chromosomes(self,
                            fitness: FitnessFunction,
                            chromosomes: list[ChromosomeType]) -> list[tuple[int | float]]:
        self._ensure_toolbox_created()
        return [fitness.evaluate(chromosome, self._toolbox.evaluate_chromosome) for chromosome in chromosomes]

    def generate_first_population(self, size_population: int) -> list[Individual]:
        self._ensure_toolbox_created()
        return self._toolbox.population(size_population)



import math
import random
from copy import deepcopy
from operator import attrgetter
from typing import Callable, Iterable

import numpy as np
from deap import base, tools
from deap.base import Toolbox

from sampo.api.genetic_api import ChromosomeType, FitnessFunction, Individual
from sampo.base import SAMPO
from sampo.scheduler.genetic.converter import (convert_schedule_to_chromosome, convert_chromosome_to_schedule,
                                               ScheduleGenerationScheme)
from sampo.scheduler.lft.base import RandomizedLFTScheduler
from sampo.scheduler.topological.base import RandomizedTopologicalScheduler
from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.resource_usage import resources_peaks_sum, resources_costs_sum, resources_sum


class TimeFitness(FitnessFunction):
    """
    Fitness function that relies on finish time.
    """

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) \
            -> tuple[int | float]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return (Time.inf().value,)
        return (schedule.execution_time.value,)


class SumOfResourcesPeaksFitness(FitnessFunction):
    """
    Fitness function that relies on sum of resources peaks usage.
    """

    def __init__(self, resources_names: Iterable[str] | None = None):
        self._resources_names = list(resources_names) if resources_names is not None else None

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) -> tuple[float]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return (Time.inf().value,)
        return (resources_peaks_sum(schedule, self._resources_names),)


class SumOfResourcesFitness(FitnessFunction):
    """
    Fitness function that relies on sum of resources usage.
    """

    def __init__(self, resources_names: Iterable[str] | None = None):
        self._resources_names = list(resources_names) if resources_names is not None else None

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) -> tuple[float]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return (Time.inf().value,)
        return (resources_sum(schedule, self._resources_names),)


class TimeWithResourcesFitness(FitnessFunction):
    """
    Fitness function that relies on finish time and the set of resources.
    """

    def __init__(self, resources_names: Iterable[str] | None = None):
        self._resources_names = list(resources_names) if resources_names is not None else None

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) -> tuple[float]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return (Time.inf().value,)
        return (schedule.execution_time.value + resources_peaks_sum(schedule, self._resources_names),)


class DeadlineResourcesFitness(FitnessFunction):
    """
    The fitness function is dependent on the set of resources and requires the end time to meet the deadline.
    """

    def __init__(self,
                 deadline: Time,
                 resources_names: Iterable[str] | None = None):
        self._deadline = deadline
        self._resources_names = list(resources_names) if resources_names is not None else None

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) \
            -> tuple[int | float]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return (Time.inf().value,)
        return (resources_peaks_sum(schedule, self._resources_names) \
                * max(1.0, schedule.execution_time.value / self._deadline.value),)


class DeadlineCostFitness(FitnessFunction):
    """
    The fitness function is dependent on the cost of resources and requires the end time to meet the deadline.
    """

    def __init__(self,
                 deadline: Time,
                 resources_names: Iterable[str] | None = None):
        self._deadline = deadline
        self._resources_names = list(resources_names) if resources_names is not None else None

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) \
            -> tuple[int | float]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return (Time.inf().value,)
        return (resources_costs_sum(schedule, self._resources_names) \
                * max(1.0, schedule.execution_time.value / self._deadline.value),)


class TimeAndResourcesFitness(FitnessFunction):
    """
    Bi-objective fitness function of finish time and sum of resources peaks.
    """

    def __init__(self, resources_names: Iterable[str] | None = None):
        self._resources_names = list(resources_names) if resources_names is not None else None

    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) \
            -> tuple[int, int]:
        schedule = evaluator(chromosome)
        if schedule is None:
            return Time.inf().value, Time.inf().value
        return schedule.execution_time.value, resources_peaks_sum(schedule, self._resources_names)


def init_toolbox(wg: WorkGraph,
                 contractors: list[Contractor],
                 worker_pool: WorkerContractorPool,
                 landscape: LandscapeConfiguration,
                 index2node: dict[int, GraphNode],
                 work_id2index: dict[str, int],
                 worker_name2index: dict[str, int],
                 index2contractor_obj: dict[int, Contractor],
                 index2zone: dict[int, str],
                 init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]],
                 mut_order_pb: float,
                 mut_res_pb: float,
                 mut_zone_pb: float,
                 statuses_available: int,
                 selection_size: int,
                 rand: random.Random,
                 spec: ScheduleSpec,
                 worker_pool_indices: dict[int, dict[int, Worker]],
                 contractor2index: dict[str, int],
                 contractor_borders: np.ndarray,
                 node_indices: list[int],
                 priorities: list[int],
                 parents: dict[int, set[int]],
                 children: dict[int, set[int]],
                 resources_border: np.ndarray,
                 assigned_parent_time: Time = Time(0),
                 fitness_weights: tuple[int | float, ...] = (-1,),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                 sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                 only_lft_initialization: bool = False,
                 is_multiobjective: bool = False) -> base.Toolbox:
    """
    Object, that include set of functions (tools) for genetic model and other functions related to it.
    list of parameters that received this function is sufficient and complete to manipulate with genetic algorithm

    :return: Object, included tools for genetic algorithm
    """
    toolbox = base.Toolbox()
    toolbox.register('register_individual_constructor', register_individual_constructor, toolbox=toolbox)
    toolbox.register_individual_constructor(fitness_weights)
    # generate chromosome
    toolbox.register('generate_chromosome', generate_chromosome, wg=wg, contractors=contractors,
                     work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=contractor2index, contractor_borders=contractor_borders, spec=spec,
                     init_chromosomes=init_chromosomes, rand=rand, work_estimator=work_estimator, landscape=landscape)

    # create population
    toolbox.register('population', generate_chromosomes, wg=wg, contractors=contractors,
                     work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=contractor2index, contractor_borders=contractor_borders, spec=spec,
                     init_chromosomes=init_chromosomes, rand=rand, work_estimator=work_estimator, landscape=landscape,
                     only_lft_initialization=only_lft_initialization, toolbox=toolbox)
    # selection
    selection = tools.selNSGA2 if is_multiobjective else select_new_population
    toolbox.register('select', selection, k=selection_size)
    # combined crossover
    toolbox.register('mate', mate, rand=rand, toolbox=toolbox, priorities=priorities)
    # combined mutation
    toolbox.register('mutate', mutate, order_mutpb=mut_order_pb, res_mutpb=mut_res_pb, zone_mutpb=mut_zone_pb,
                     rand=rand, parents=parents, children=children, resources_border=resources_border,
                     statuses_available=statuses_available, priorities=priorities)
    # crossover for order
    toolbox.register('mate_order', mate_scheduling_order, rand=rand, toolbox=toolbox, priorities=priorities)
    # mutation for order
    toolbox.register('mutate_order', mutate_scheduling_order, mutpb=mut_order_pb, rand=rand, parents=parents,
                     children=children, priorities=priorities)
    # crossover for resources
    toolbox.register('mate_resources', mate_resources, rand=rand, toolbox=toolbox)
    # mutation for resources
    toolbox.register('mutate_resources', mutate_resources, resources_border=resources_border,
                     mutpb=mut_res_pb, rand=rand)
    # mutation for resource borders
    toolbox.register('mutate_resource_borders', mutate_resource_borders, contractor_borders=contractor_borders,
                     mutpb=mut_res_pb, rand=rand)
    toolbox.register('mate_post_zones', mate_for_zones, rand=rand, toolbox=toolbox)
    toolbox.register('mutate_post_zones', mutate_for_zones, rand=rand, mutpb=mut_zone_pb,
                     statuses_available=landscape.zone_config.statuses.statuses_available())

    toolbox.register('validate', is_chromosome_correct, node_indices=node_indices, parents=parents,
                     contractor_borders=contractor_borders, index2node=index2node)
    toolbox.register('schedule_to_chromosome', convert_schedule_to_chromosome,
                     work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=contractor2index, contractor_borders=contractor_borders, spec=spec,
                     landscape=landscape)
    toolbox.register('evaluate_chromosome', evaluate, wg=wg, toolbox=toolbox)
    toolbox.register('chromosome_to_schedule', convert_chromosome_to_schedule, worker_pool=worker_pool,
                     index2node=index2node, index2contractor=index2contractor_obj,
                     worker_pool_indices=worker_pool_indices, assigned_parent_time=assigned_parent_time,
                     work_estimator=work_estimator, worker_name2index=worker_name2index,
                     contractor2index=contractor2index, index2zone=index2zone,
                     landscape=landscape, sgs_type=sgs_type)
    toolbox.register('copy_individual', copy_individual, toolbox=toolbox)

    return toolbox


def evaluate(chromosome: ChromosomeType, wg: WorkGraph, toolbox: Toolbox) -> Schedule | None:
    if toolbox.validate(chromosome):
        sworks = toolbox.chromosome_to_schedule(chromosome)[0]
        return Schedule.from_scheduled_works(sworks.values(), wg)
    else:
        return None


def register_individual_constructor(fitness_weights: tuple[int | float, ...], toolbox: base.Toolbox):
    class IndividualFitness(base.Fitness):
        weights = fitness_weights

    toolbox.register('Individual', Individual.prepare(IndividualFitness))


def copy_individual(ind: Individual, toolbox: Toolbox) -> Individual:
    return toolbox.Individual(
        (ind[0].copy(), ind[1].copy(), ind[2].copy(), deepcopy(ind[3]), ind[4].copy())
    )


def generate_chromosomes(n: int,
                         wg: WorkGraph,
                         contractors: list[Contractor],
                         spec: ScheduleSpec,
                         work_id2index: dict[str, int],
                         worker_name2index: dict[str, int],
                         contractor2index: dict[str, int],
                         contractor_borders: np.ndarray,
                         init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]],
                         rand: random.Random,
                         toolbox: Toolbox,
                         work_estimator: WorkTimeEstimator = None,
                         landscape: LandscapeConfiguration = LandscapeConfiguration(),
                         only_lft_initialization: bool = False) -> list[ChromosomeType]:
    """
    Generates n chromosomes.
    Do not use `generate_chromosome` function.
    """

    def randomized_init(is_topological: bool = False) -> ChromosomeType:
        if is_topological:
            seed = int(rand.random() * 1000000)
            schedule, _, _, node_order = RandomizedTopologicalScheduler(work_estimator,
                                                                        seed).schedule_with_cache(wg, contractors, spec,
                                                                                                  landscape=landscape)[0]
        else:
            schedule, _, _, node_order = RandomizedLFTScheduler(work_estimator=work_estimator,
                                                                rand=rand).schedule_with_cache(wg, contractors, spec,
                                                                                               landscape=landscape)[0]
        return convert_schedule_to_chromosome(work_id2index, worker_name2index, contractor2index, contractor_borders,
                                              schedule, spec, landscape, node_order)

    if only_lft_initialization:
        chromosomes = [toolbox.Individual(randomized_init(is_topological=False)) for _ in range(n - 1)]
        chromosomes.append(toolbox.Individual(init_chromosomes['lft'][0]))
        return chromosomes

    count_for_specified_types = (n // 3) // len(init_chromosomes)
    count_for_specified_types = count_for_specified_types if count_for_specified_types > 0 else 1
    weights = [importance for _, importance, _ in init_chromosomes.values()]
    sum_of_weights = sum(weights)
    weights = [weight / sum_of_weights for weight in weights]

    counts = [math.ceil(count_for_specified_types * weight) for weight in weights]
    sum_counts_for_specified_types = sum(counts)

    count_for_topological = n // 2 - sum_counts_for_specified_types
    count_for_topological = count_for_topological if count_for_topological > 0 else 1
    counts.append(count_for_topological)

    count_for_rand_lft = n - count_for_topological - sum_counts_for_specified_types
    count_for_rand_lft = count_for_rand_lft if count_for_rand_lft > 0 else 1
    counts.append(count_for_rand_lft)

    chromosome_types = rand.sample(list(init_chromosomes.keys()) + ['topological', 'rand_lft'], k=n, counts=counts)

    chromosomes = []

    for generated_type in chromosome_types:
        match generated_type:
            case 'topological':
                ind = randomized_init(is_topological=True)
            case 'rand_lft':
                ind = randomized_init(is_topological=False)
            case _:
                ind = init_chromosomes[generated_type][0]

        if not toolbox.validate(ind):
            SAMPO.logger.warn('HELP')

        ind = toolbox.Individual(ind)
        chromosomes.append(ind)

    return chromosomes[:n]


def generate_chromosome(wg: WorkGraph,
                        contractors: list[Contractor],
                        work_id2index: dict[str, int],
                        worker_name2index: dict[str, int],
                        contractor2index: dict[str, int],
                        contractor_borders: np.ndarray,
                        init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]],
                        spec: ScheduleSpec,
                        rand: random.Random,
                        work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                        landscape: LandscapeConfiguration = LandscapeConfiguration()) -> ChromosomeType:
    """
    It is necessary to generate valid scheduling, which are satisfied to current dependencies
    That's why will be used the approved order of works (HEFT order and Topological sorting)
    Topological sorts are generating always different
    HEFT is always the same(now not)
    HEFT we will choose in 30% of attempts
    Topological in others
    """

    def randomized_init() -> ChromosomeType:
        schedule, _, _, node_order = RandomizedTopologicalScheduler(work_estimator,
                                                  int(rand.random() * 1000000)) \
            .schedule_with_cache(wg, contractors, spec, landscape=landscape)[0]
        return convert_schedule_to_chromosome(work_id2index, worker_name2index,
                                              contractor2index, contractor_borders,
                                              schedule, spec, landscape, node_order)

    chance = rand.random()
    if chance < 0.2:
        chromosome = init_chromosomes['heft_end'][0]
    elif chance < 0.4:
        chromosome = init_chromosomes['heft_between'][0]
    elif chance < 0.5:
        chromosome = init_chromosomes['12.5%'][0]
    elif chance < 0.6:
        chromosome = init_chromosomes['25%'][0]
    elif chance < 0.7:
        chromosome = init_chromosomes['75%'][0]
    elif chance < 0.8:
        chromosome = init_chromosomes['87.5%'][0]
    else:
        chromosome = randomized_init()

    return chromosome


def select_new_population(population: list[Individual], k: int) -> list[Individual]:
    """
    Selection operator for genetic algorithm.
    Select top k individuals in population.
    """
    population = sorted(population, key=attrgetter('fitness'), reverse=True)
    return population[:k]


def is_chromosome_correct(ind: Individual, node_indices: list[int], parents: dict[int, set[int]],
                          contractor_borders: np.ndarray, index2node: dict[int, GraphNode]) -> bool:
    """
    Check correctness of works order and contractors borders.
    """
    return is_chromosome_order_correct(ind, parents, index2node) and \
        is_chromosome_contractors_correct(ind, node_indices, contractor_borders)


def is_chromosome_order_correct(ind: Individual, parents: dict[int, set[int]], index2node: dict[int, GraphNode]) -> bool:
    """
    Checks that assigned order of works are topologically correct.
    """
    work_order = ind[0]
    used = set()
    for work_index in work_order:
        used.add(work_index)
        if not parents[work_index].issubset(used):
            # logger.error(f'Order validation failed: {work_order}')
            return False

        # validate priorities
        work_node = index2node[work_index]
        if any(index2node[parent].work_unit.priority > work_node.work_unit.priority for parent in parents[work_index]):
            # SAMPO.logger.error(f'Order validation failed')
            return False
    return True


def is_chromosome_contractors_correct(ind: Individual, work_indices: Iterable[int],
                                      contractor_borders: np.ndarray) -> bool:
    """
    Checks that assigned contractors can supply assigned workers.
    """
    if not work_indices:
        return True
    resources = ind[1][work_indices]
    # sort resource part of chromosome by contractor ids
    resources = resources[resources[:, -1].argsort()]
    # get unique contractors and indexes where they start
    contractors, indexes = np.unique(resources[:, -1], return_index=True)
    # get borders of received contractors from chromosome
    chromosome_borders = ind[2][contractors]
    # split resources to get parts grouped by contractor parts
    res_grouped_by_contractor = np.split(resources[:, :-1], indexes[1:])
    # for each grouped parts take maximum for each resource
    max_of_res_by_contractor = np.array([r.max(axis=0) for r in res_grouped_by_contractor])
    return (max_of_res_by_contractor <= chromosome_borders).all() and \
        (chromosome_borders <= contractor_borders[contractors]).all()


def get_order_part(order: np.ndarray, other_order: np.ndarray) -> np.ndarray:
    """
    Get a new part in topologic order for chromosome.
    This function is needed to make crossover for order.
    """
    order = set(order)
    return np.array([node for node in other_order if node not in order])


def mate_scheduling_order(ind1: Individual, ind2: Individual, rand: random.Random,
                          toolbox: Toolbox, priorities: np.ndarray, copy: bool = True) -> tuple[Individual, Individual]:
    """
    Two-Point crossover for order.

    :param ind1: first individual
    :param ind2: second individual
    :param rand: the rand object used for randomized operations
    :param toolbox: toolbox
    :param priorities: node priorities
    :param copy: if True individuals will be copied before mating so as not to change them

    :return: two mated individuals
    """
    child1, child2 = (toolbox.copy_individual(ind1), toolbox.copy_individual(ind2)) if copy else (ind1, ind2)

    def mate_parts(part1, part2):
        parent1 = part1.copy()

        min_mating_amount = len(part1) // 4

        two_point_order_crossover(part1, part2, min_mating_amount, rand)
        two_point_order_crossover(part2, parent1, min_mating_amount, rand)

    order1, order2 = child1[0], child2[0]

    # mate parts inside priority groups

    # priorities of tasks with same order-index should be the same (if chromosome is valid)
    cur_priority = priorities[order1[0]]
    cur_priority_group_start = 0
    for i in range(len(order1)):
        if priorities[order1[i]] != cur_priority:
            cur_priority = priorities[order1[i]]

            mate_parts(order1[cur_priority_group_start:i], order2[cur_priority_group_start:i])

    return toolbox.Individual(child1), toolbox.Individual(child2)


def two_point_order_crossover(child: np.ndarray, other_parent: np.ndarray, min_mating_amount: int, rand: random.Random):
    """
    This function implements Two-Point crossover for order.

    :param child: order to which implements crossover, it is equal to order of first parent.
    :param other_parent: order of second parent from which mating part will be taken.
    :param min_mating_amount: minimum amount of mating part
    :param rand: the rand object used for randomized operations

    :return: child mated with other parent
    """
    # randomly select mating amount for child
    mating_amount = rand.randint(min_mating_amount, 3 * min_mating_amount)
    if mating_amount > 1:
        # based on received mating amount randomly select the points between which the crossover will take place
        crossover_head_point = rand.randint(1, mating_amount - 1)
        crossover_tail_point = mating_amount - crossover_head_point

        # get mating order part from parent
        ind_new_part = get_order_part(np.concatenate((child[:crossover_head_point], child[-crossover_tail_point:])),
                                      other_parent)
        # update mating part to received values
        child[crossover_head_point:-crossover_tail_point] = ind_new_part

    return child


def mutate_scheduling_order_core(order: np.ndarray, mutpb: float, rand: random.Random,
                                 parents: dict[int, set[int]], children: dict[int, set[int]]):
    # number of possible mutations = number of works except start and finish works
    num_possible_muts = len(order) - 2
    # generate mask of works to mutate based on mutation probability
    mask = np.array([rand.random() < mutpb for _ in range(num_possible_muts)])
    if mask.any():
        # get indexes of works to mutate based on generated mask
        # +1 because start work was not taken into account in mask generation
        indexes_of_works_to_mutate = np.where(mask)[0] + 1
        # shuffle order of mutations
        rand.shuffle(indexes_of_works_to_mutate)
        # get works to mutate based on shuffled indexes
        works_to_mutate = order[indexes_of_works_to_mutate]
        for work in works_to_mutate:
            # pop index of the current work
            i, indexes_of_works_to_mutate = indexes_of_works_to_mutate[0], indexes_of_works_to_mutate[1:]
            # find max index of parent of the current work
            # +1 because insertion should be righter
            i_parent = np.max(np.where(np.isin(order[:i], list(parents[work]), assume_unique=True))[0],
                              initial=0) + 1
            # find min index of child of the current work
            # +i because the slice [i + 1:] was taken, and +1 is not needed because these indexes will be shifted left
            # after current work deletion
            i_children = np.min(np.where(np.isin(order[i + 1:], list(children[work]), assume_unique=True))[0],
                                initial=len(order) - 2 - i) + i
            if i_parent == i_children:
                # if child and parent indexes are equal then no mutation can be done
                continue
            else:
                # shift work indexes (which are to the right of the current index) to the left
                # after the current work deletion
                indexes_of_works_to_mutate[indexes_of_works_to_mutate > i] -= 1
                # range potential indexes to insert the current work
                choices = np.concatenate((np.arange(i_parent, i), np.arange(i + 1, i_children + 1)))
                # set weights to potential indexes based on their distance from the current one
                weights = 1 / np.abs(choices - i)
                # generate new index for the current work
                new_i = rand.choices(choices, weights=weights)[0]
                # delete current work from current index, insert in new generated index and update scheduling order
                # in chromosome
                order[:] = np.insert(np.delete(order, i), new_i, work)
                # shift work indexes (which are to the right or equal to the new index) to the right
                # after the current work insertion in new generated index
                indexes_of_works_to_mutate[indexes_of_works_to_mutate >= new_i] += 1


def mutate_scheduling_order(ind: Individual, mutpb: float, rand: random.Random, priorities: np.ndarray,
                            parents: dict[int, set[int]], children: dict[int, set[int]]) -> Individual:
    """
    Mutation operator for works scheduling order.

    :param ind: the individual to be mutated
    :param mutpb: probability of gene mutation
    :param rand: the rand object used for randomized operations
    :param priorities: node priorities
    :param parents: mapping object of works and their parent-works to create valid order
    :param children: mapping object of works and their children-works to create valid order

    :return: mutated individual
    """
    order = ind[0]

    priority_groups_count = len(set(priorities))
    mutpb_for_priority_group = mutpb #/ priority_groups_count

    # priorities of tasks with same order-index should be the same (if chromosome is valid)
    cur_priority = priorities[order[0]]
    cur_priority_group_start = 0
    for i in range(len(order)):
        if priorities[order[i]] != cur_priority:
            mutate_scheduling_order_core(order[cur_priority_group_start:i],
                                         mutpb_for_priority_group,
                                         rand, parents, children)

            cur_priority = priorities[order[i]]
            cur_priority_group_start = i

    return ind


def mate_resources(ind1: Individual, ind2: Individual, rand: random.Random,
                   optimize_resources: bool, toolbox: Toolbox, copy: bool = True) -> tuple[Individual, Individual]:
    """
    One-Point crossover for resources.

    :param ind1: first individual
    :param ind2: second individual
    :param optimize_resources: if True resource borders should be changed after mating
    :param rand: the rand object used for randomized operations
    :param copy: if True individuals will be copied before mating so as not to change them
    :param toolbox: toolbox

    :return: two mated individuals
    """
    child1, child2 = (toolbox.copy_individual(ind1), toolbox.copy_individual(ind2)) if copy else (ind1, ind2)

    res1, res2 = child1[1], child2[1]
    num_works = len(res1)
    min_mating_amount = num_works // 4
    cxpoint = rand.randint(min_mating_amount, num_works - min_mating_amount)
    mate_positions = rand.sample(range(num_works), cxpoint)

    res1[mate_positions], res2[mate_positions] = res2[mate_positions], res1[mate_positions]

    if optimize_resources:
        for res, child in zip([res1, res2], [child1, child2]):
            mated_resources = res[mate_positions]
            # take contractors from mated positions
            contractors = np.unique(mated_resources[:, -1])
            # take maximum from borders of these contractors in two chromosomes to maintain validity
            # and update current child borders on received maximum
            child[2][contractors] = np.stack((child1[2][contractors], child2[2][contractors]), axis=0).max(axis=0)

    return toolbox.Individual(child1), toolbox.Individual(child2)


def mutate_resources(ind: Individual, mutpb: float, rand: random.Random,
                     resources_border: np.ndarray) -> Individual:
    """
    Mutation function for resources.
    It changes selected numbers of workers in random work in a certain interval for this work.

    :param ind: the individual to be mutated
    :param resources_border: low and up borders of resources amounts
    :param mutpb: probability of gene mutation
    :param rand: the rand object used for randomized operations

    :return: mutated individual
    """
    res = ind[1]
    num_works = len(res)

    num_contractors = len(ind[2])
    if num_contractors > 1:
        mask = np.array([rand.random() < mutpb for _ in range(num_works)])
        if mask.any():
            # generate new contractors in the number of received True values of mask
            new_contractors = np.array([rand.randint(0, num_contractors - 1) for _ in range(mask.sum())])
            # obtain a new mask of correspondence
            # between the borders of the received contractors and the assigned resources
            contractor_mask = (res[mask, :-1] <= ind[2][new_contractors]).all(axis=1)
            # update contractors by received mask
            new_contractors = new_contractors[contractor_mask]
            # update mask by new mask
            mask[mask] &= contractor_mask
            # mutate contractors
            res[mask, -1] = new_contractors

    num_res = len(res[0, :-1])
    res_indexes = np.arange(0, num_res)
    works_indexes = np.arange(0, num_works)
    masks = np.array([[rand.random() < mutpb for _ in range(num_res)] for _ in range(num_works)])
    # mask of works where at least one resource should be mutated
    mask = masks.any(axis=1)

    if not mask.any():
        # if no True value in mask then no mutation can be done
        return ind

    # get works indexes where mutation should be done and their masks of resources to be mutated
    works_indexes, masks = works_indexes[mask], masks[mask]
    # get up borders of resources of works where mutation should be done
    # by taking minimum (borders of the contractors assigned to them) and (maximum values of resources for these works)
    res_up_borders = np.stack((resources_border[1].T[mask], ind[2][res[mask, -1]]), axis=0).min(axis=0)
    # get minimum values of resources for these works
    res_low_borders = resources_border[0].T[mask]
    # if low border and up border are equal then no mutation can be done
    # update masks by checking this condition
    masks &= res_up_borders != res_low_borders
    # update mask of works where mutation should be done
    mask = masks.any(axis=1)

    # make mutation of resources
    mutate_values(res, works_indexes[mask], res_indexes, res_low_borders[mask],
                  res_up_borders[mask], masks[mask], -1, rand)

    return ind


def mate(ind1: Individual, ind2: Individual, optimize_resources: bool,
         rand: random.Random, toolbox: Toolbox, priorities: np.ndarray) \
        -> tuple[Individual, Individual]:
    """
    Combined crossover function of Two-Point crossover for order, One-Point crossover for resources
    and One-Point crossover for zones.

    :param ind1: first individual
    :param ind2: second individual
    :param optimize_resources: if True resource borders should be changed after mating
    :param rand: the rand object used for randomized operations
    :param priorities: priorities
    :param toolbox: toolbox

    :return: two mated individuals
    """
    child1, child2 = mate_scheduling_order(ind1, ind2, rand, toolbox, priorities, copy=True)
    child1, child2 = mate_resources(child1, child2, rand, optimize_resources, toolbox, copy=False)
    # TODO Make better crossover for zones and uncomment this
    # child1, child2 = mate_for_zones(child1, child2, rand, copy=False)

    return toolbox.Individual(child1), toolbox.Individual(child2)


def mutate(ind: Individual, resources_border: np.ndarray, parents: dict[int, set[int]],
           children: dict[int, set[int]], statuses_available: int, priorities: np.ndarray,
           order_mutpb: float, res_mutpb: float, zone_mutpb: float,
           rand: random.Random) -> Individual:
    """
    Combined mutation function of mutation for order, mutation for resources and mutation for zones.

    :param ind: the individual to be mutated
    :param resources_border: low and up borders of resources amounts
    :param parents: mapping object of works and their parent-works to create valid order
    :param children: mapping object of works and their children-works to create valid order
    :param statuses_available: number of statuses available
    :param order_mutpb: probability of order's gene mutation
    :param res_mutpb: probability of resources' gene mutation
    :param zone_mutpb: probability of zones' gene mutation
    :param rand: the rand object used for randomized operations

    :return: mutated individual
    """
    mutant = mutate_scheduling_order(ind, order_mutpb, rand, priorities, parents, children)
    mutant = mutate_resources(mutant, res_mutpb, rand, resources_border)
    # TODO Make better mutation for zones and uncomment this
    # mutant = mutate_for_zones(mutant, statuses_available, zone_mutpb, rand)

    return mutant


def mutate_resource_borders(ind: Individual, mutpb: float, rand: random.Random,
                            contractor_borders: np.ndarray) -> Individual:
    """
    Mutation function for contractors' resource borders.

    :param ind: the individual to be mutated
    :param contractor_borders: up borders of contractors capacity
    :param mutpb: probability of gene mutation
    :param rand: the rand object used for randomized operations

    :return: mutated individual
    """
    borders = ind[2]
    res = ind[1]
    num_res = len(res[0, :-1])
    res_indexes = np.arange(0, num_res)
    # sort resource part of chromosome by contractor ids
    resources = res[res[:, -1].argsort()]
    # get unique contractors and indexes where they start
    contractors, indexes = np.unique(resources[:, -1], return_index=True)
    # split resources to get parts grouped by contractor parts
    res_grouped_by_contractor = np.split(resources[:, :-1], indexes[1:])
    masks = np.array([[rand.random() < mutpb for _ in range(num_res)] for _ in contractors])
    # mask of contractors where at least one resource border should be mutated
    mask = masks.any(axis=1)

    if not mask.any():
        # if no True value in mask then no mutation can be done
        return ind

    # get contractors where mutation should be done and their masks of resource borders to be mutated
    contractors, masks = contractors[mask], masks[mask]
    # get maximum values of resource borders for received contractors
    contractor_up_borders = contractor_borders[contractors]
    # get minimum values of resource borders of contractors where mutation should be done
    # by taking maximum of assigned resources for works which have contractor that should be mutated
    contractor_low_borders = np.array([r.max(axis=0) for r, is_mut in zip(res_grouped_by_contractor, mask) if is_mut])
    # if minimum and maximum values are equal then no mutation can be done
    # update masks by checking this condition
    masks &= contractor_up_borders != contractor_low_borders
    # update mask of contractors where mutation should be done
    mask = masks.any(axis=1)

    # make mutation of resource borders
    mutate_values(borders, contractors[mask], res_indexes,
                  contractor_low_borders[mask], contractor_up_borders[mask],
                  masks[mask], len(res_indexes), rand)

    return ind


def mutate_values(chromosome_part: np.ndarray, row_indexes: np.ndarray, col_indexes: np.ndarray,
                  low_borders: np.ndarray, up_borders: np.ndarray, masks: np.ndarray, mut_part: int,
                  rand: random.Random) -> None:
    """
    Changes numeric values in m x n part of chromosome.
    This function is needed to make mutation for resources and resource borders.
    """
    for row_index, l_borders, u_borders, row_mask in zip(row_indexes, low_borders, up_borders, masks):
        cur_row = chromosome_part[row_index]
        for col_index, current_amount, l_border, u_border in zip(col_indexes[row_mask], cur_row[:mut_part][row_mask],
                                                                 l_borders[row_mask], u_borders[row_mask]):
            # range new potential amount except current amount
            choices = np.concatenate((np.arange(l_border, current_amount),
                                      np.arange(current_amount + 1, u_border + 1)))
            # set weights to potential amounts based on their distance from the current one
            weights = 1 / np.abs(choices - current_amount)

            assert l_border <= u_border, f'{l_border} {u_border} {current_amount}'
            assert len(weights) > 0

            cur_row[col_index] = rand.choices(choices, weights=weights)[0]


def mate_for_zones(ind1: Individual, ind2: Individual, rand: random.Random,
                   toolbox: Toolbox, copy: bool = True) -> tuple[Individual, Individual]:
    """
    CxOnePoint for zones

    :param ind1: first individual
    :param ind2: second individual
    :param rand: the rand object used for randomized operations
    :param copy: if True individuals will be copied before mating so as not to change them
    :param toolbox: toolbox

    :return: two mated individuals
    """
    child1, child2 = (toolbox.copy_individual(ind1), toolbox.copy_individual(ind2)) if copy else (ind1, ind2)

    zones1 = child1[4]
    zones2 = child2[4]

    if zones1.size:
        num_works = len(zones1)
        border = num_works // 4
        cxpoint = rand.randint(border, num_works - border)

        mate_positions = rand.sample(range(num_works), cxpoint)

        zones1[mate_positions], zones2[mate_positions] = zones2[mate_positions], zones1[mate_positions]

    return toolbox.Individual(child1), toolbox.Individual(child2)


def mutate_for_zones(ind: Individual, mutpb: float, rand: random.Random, statuses_available: int) -> Individual:
    """
    Mutation function for zones.
    It changes selected numbers of zones in random work in a certain interval from available statuses.

    :param ind: the individual to be mutated
    :param mutpb: probability of gene mutation
    :param rand: the rand object used for randomized operations
    :param statuses_available: number of statuses available

    :return: mutated individual
    """
    # select random number from interval from min to max from uniform distribution
    zones = ind[4]
    if zones.size:
        mask = np.array([[rand.random() < mutpb for _ in range(zones.shape[1])] for _ in range(zones.shape[0])])
        new_zones = np.array([rand.randint(0, statuses_available - 1) for _ in range(mask.sum())])
        zones[mask] = new_zones

    return ind



from dataclasses import dataclass
from typing import Any, Callable

from sampo.schemas.contractor import Contractor
from sampo.schemas.landscape import MaterialDelivery
from sampo.schemas.resources import Equipment, ConstructionObject, Worker
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.works import WorkUnit
from sampo.schemas.zones import ZoneTransition
from sampo.utilities.serializers import custom_serializer


@dataclass
class ScheduledWork(AutoJSONSerializable['ScheduledWork']):
    """
    Contains all necessary info to represent WorkUnit in schedule:

    * WorkUnit
    * list of workers, that are required to complete task
    * start and end time
    * contractor, that complete task
    * list of equipment, that is needed to complete the task
    * list of materials - set of non-renewable resources
    * object - variable, that is used in landscape
    """

    ignored_fields = ['equipments', 'materials', 'object', 'work_unit']

    def __init__(self,
                 work_unit: WorkUnit,
                 start_end_time: tuple[Time, Time],
                 workers: list[Worker],
                 contractor: Contractor | str,
                 equipments: list[Equipment] | None = None,
                 zones_pre: list[ZoneTransition] | None = None,
                 zones_post: list[ZoneTransition] | None = None,
                 materials: MaterialDelivery | None = None,
                 c_object: ConstructionObject | None = None):
        self.id = work_unit.id
        self.name = work_unit.name
        self.display_name = work_unit.display_name
        self.is_service_unit = work_unit.is_service_unit
        self.volume = work_unit.volume
        self.volume_type = work_unit.volume_type
        self.priority = work_unit.priority
        self.start_end_time = start_end_time
        self.workers = workers if workers is not None else []
        self.equipments = equipments if equipments is not None else []
        self.zones_pre = zones_pre if zones_pre is not None else []
        self.zones_post = zones_post if zones_post is not None else []
        self.materials = materials if materials is not None else MaterialDelivery('')
        self.object = c_object if c_object is not None else []

        if contractor is not None:
            if isinstance(contractor, str):
                self.contractor = contractor
            else:
                self.contractor = contractor.name if contractor.name else contractor.id
        else:
            self.contractor = ""

        self.cost = sum([worker.get_cost() * self.duration.value for worker in self.workers])

    def __str__(self) -> str:
        return f'ScheduledWork[work_unit={self.id}, start_end_time={self.start_end_time}, ' \
               f'workers={self.workers}, contractor={self.contractor}]'

    def __repr__(self) -> str:
        return self.__str__()

    @custom_serializer('workers')
    @custom_serializer('zones_pre')
    @custom_serializer('zones_post')
    @custom_serializer('start_end_time')
    def serialize_serializable_list(self, value) -> list:
        return [t._serialize() for t in value]

    @classmethod
    @custom_serializer('start_end_time', deserializer=True)
    def deserialize_time(cls, value) -> tuple[Time, Time]:
        return Time._deserialize(value[0]), Time._deserialize(value[1])

    @classmethod
    @custom_serializer('workers', deserializer=True)
    def deserialize_workers(cls, value) -> list[Worker]:
        return [Worker._deserialize(t) for t in value]

    @classmethod
    @custom_serializer('zones_pre', deserializer=True)
    @custom_serializer('zones_post', deserializer=True)
    def deserialize_zone_transitions(cls, value) -> list[ZoneTransition]:
        return [ZoneTransition._deserialize(t) for t in value]

    @property
    def start_time(self) -> Time:
        return self.start_end_time[0]

    @start_time.setter
    def start_time(self, val: Time):
        self.start_end_time = (val, self.start_end_time[1])

    @property
    def finish_time(self) -> Time:
        return self.start_end_time[1]

    @finish_time.setter
    def finish_time(self, val: Time):
        self.start_end_time = (self.start_end_time[0], val)

    @staticmethod
    def start_time_getter() -> Callable[[], Time]:
        return lambda x: x.start_end_time[0]

    @staticmethod
    def finish_time_getter() -> Callable[[], Time]:
        return lambda x: x.start_end_time[1]

    @property
    def duration(self) -> Time:
        start, end = self.start_end_time
        return end - start

    def to_dict(self) -> dict[str, Any]:
        return {
            'task_id': self.id,
            'task_name': self.name,
            'start': self.start_time.value,
            'finish': self.finish_time.value,
            'contractor_id': self.contractor,
            'priority': self.priority,
            'workers': {worker.name: worker.count for worker in self.workers},
        }



import datetime
import math
from typing import Tuple

import numpy as np
import pandas as pd

from sampo.schemas.graph import EdgeType
from sampo.utilities.name_mapper import NameMapper


def get_all_connections(graph_df: pd.DataFrame,
                        use_mapper: bool = False,
                        mapper: NameMapper | None = None) \
        -> Tuple[dict[str, list], dict[str, list]]:

    task_name_column = 'granular_name'

    num_tasks = len(graph_df)
    # Get the upper triangular indices to avoid duplicate pairs
    indices = np.triu_indices(num_tasks, k=1)
    works1_ids = graph_df['activity_id'].values[indices[0]]
    works1_names = graph_df[task_name_column].values[indices[0]]

    works2_ids = graph_df['activity_id'].values[indices[1]]
    works2_names = graph_df[task_name_column].values[indices[1]]

    if use_mapper:
        works1_names = np.vectorize(mapper.get)(works1_names)
        works2_names = np.vectorize(mapper.get)(works2_names)

    return {"ids": works1_ids, "names": works1_names}, {"ids": works2_ids, "names": works2_names}


def get_delta_between_dates(first: str, second: str) -> int:
    return max((datetime.date(int(first.split('-')[0]), int(first.split('-')[1]), int(first.split('-')[2])) -
                datetime.date(int(second.split('-')[0]), int(second.split('-')[1]), int(second.split('-')[2]))).days, 1)


def find_min_without_outliers(lst: list[float]) -> float:
    return round(min([x for x in lst if x >= np.mean(lst) - 3 * np.std(lst)]), 2)


def gather_links_types_statistics(s1: str, f1: str, s2: str, f2: str) \
        -> Tuple[int, int, int, list, list, int, list, list, int, list, list, int, list, list]:
    """
    Count statistics on the occurrence of different mutual arrangement of tasks

    :param s1: start of first work
    :param f1: finish of first work
    :param s2: start of second work
    :param f2: finish of second work
    :return: Statistics on the occurrence of different mutual arrangement of tasks
    """

    fs12, fs21, ss12, ss21 = 0, 0, 0, 0
    ss12_lags, ss12_percent_lags, ss21_lags, ss21_percent_lags = [], [], [], []

    ffs12, ffs21 = 0, 0
    ffs12_lags, ffs12_percent_lags, ffs21_lags, ffs21_percent_lags = [], [], [], []

    if s1 == s2 and f1 == f2:
        ffs12 += 1
        ffs12_percent_lags.append(0.01)
        ffs12_lags.append(0.01)
    if f2 <= s1:
        fs21 += 1
    else:
        if s2 >= f1:
            fs12 += 1
        else:
            if s2 >= s1:
                if f2 >= f1:
                    ffs12 += 1
                    if get_delta_between_dates(f1, s1) != 0:
                        ffs12_percent_lags.append(get_delta_between_dates(s2, s1) / get_delta_between_dates(f1, s1))
                    else:
                        ffs12_percent_lags.append(0)
                    ffs12_lags.append(get_delta_between_dates(s2, s1))
                else:
                    ss12 += 1
                    if get_delta_between_dates(f1, s1) != 0:
                        ss12_percent_lags.append(get_delta_between_dates(s2, s1) / get_delta_between_dates(f1, s1))
                    else:
                        ss12_percent_lags.append(0)
                    ss12_lags.append(get_delta_between_dates(s2, s1))
            else:
                if f2 <= f1:
                    ffs21 += 1
                    if get_delta_between_dates(f2, s2) != 0:
                        ffs21_percent_lags.append(get_delta_between_dates(s1, s2) / get_delta_between_dates(f2, s2))
                    else:
                        ffs21_percent_lags.append(0)
                    ffs21_lags.append(get_delta_between_dates(s1, s2))
                else:
                    ss21 += 1
                    if get_delta_between_dates(f2, s2) != 0:
                        ss21_percent_lags.append(get_delta_between_dates(s1, s2) / get_delta_between_dates(f2, s2))
                    else:
                        ss21_percent_lags.append(0)
                    ss21_lags.append(get_delta_between_dates(s1, s2))
    return fs12, fs21, ss12, ss12_lags, ss12_percent_lags, ss21, ss21_lags, ss21_percent_lags, ffs12, ffs12_lags, ffs12_percent_lags, \
        ffs21, ffs21_lags, ffs21_percent_lags


def get_all_seq_statistic(history_data: pd.DataFrame,
                          graph_df: pd.DataFrame,
                          use_model_name: bool = False,
                          mapper: NameMapper | None = None):
    df_grouped = history_data.copy()

    if use_model_name:
        column_name = 'model_name'
    else:
        if 'granular_name' not in history_data.columns:
            history_data['granular_name'] = [activity_name for activity_name in history_data['work_name']]
        column_name = 'granular_name'

    df_grouped = df_grouped.groupby('upper_works')[column_name].apply(list).reset_index(name="Works")
    works1, works2 = get_all_connections(graph_df, use_model_name, mapper)

    # Declare structure with updated connections

    tasks_names = list(zip(works1['names'], works2['names']))
    tasks_ids = list(zip(works1['ids'], works2['ids']))

    predecessors_info_dict = {w_id: [] for w_id in graph_df['activity_id']}

    if len(tasks_names) != 0:
        for i in range(len(tasks_names)):
            w1, w2 = tasks_names[i]
            w1_id, w2_id = tasks_ids[i]

            if w1 != w2:
                fs12, fs21, ss12, ss21 = 0, 0, 0, 0
                ss12_lags, ss12_percent_lags, ss21_lags, ss21_percent_lags = [], [], [], []

                count = 0

                ffs12, ffs21 = 0, 0
                ffs12_lags, ffs12_percent_lags, ffs21_lags, ffs21_percent_lags = [], [], [], []

                for i, work_list in df_grouped.iterrows():
                    # Looking to see if this pair of works occurred within the same site in the historical data
                    if w1 in work_list['Works'] and w2 in work_list['Works']:
                        ind1 = history_data.loc[(history_data['upper_works'] == work_list['upper_works']) &
                                                (history_data[column_name] == w1)]
                        ind2 = history_data.loc[(history_data['upper_works'] == work_list['upper_works']) &
                                                (history_data[column_name] == w2)]

                        ind1_sorted = ind1.sort_values(by=['first_day', 'last_day']).reset_index(drop=True)
                        ind2_sorted = ind2.sort_values(by=['first_day', 'last_day']).reset_index(drop=True)

                        for l in range(min(len(ind1_sorted), len(ind2_sorted))):
                            s1, f1 = ind1_sorted.loc[l, 'first_day'], ind1_sorted.loc[l, 'last_day']

                            s2, f2 = ind2_sorted.loc[l, 'first_day'], ind2_sorted.loc[l, 'last_day']

                            if not any([type(x) == float for x in [s1, s2, f1, f2]]):
                                tasks_fs12, tasks_fs21, tasks_ss12, tasks_ss12_lags, tasks_ss12_percent_lags, tasks_ss21, tasks_ss21_lags, \
                                    tasks_ss21_percent_lags, tasks_ffs12, tasks_ffs12_lags, tasks_ffs12_percent_lags, tasks_ffs21, tasks_ffs21_lags, tasks_ffs21_percent_lags = gather_links_types_statistics(
                                    s1, f1, s2, f2)

                                count += 1

                                fs12 += tasks_fs12
                                fs21 += tasks_fs21

                                ss12 += tasks_ss12
                                ss12_lags.extend(tasks_ss12_lags)
                                ss12_percent_lags.extend(tasks_ss12_percent_lags)
                                ss21 += tasks_ss21
                                ss21_lags.extend(tasks_ss21_lags)
                                ss21_percent_lags.extend(tasks_ss21_percent_lags)

                                ffs12 += tasks_ffs12
                                ffs12_lags.extend(tasks_ffs12_lags)
                                ffs12_percent_lags.extend(tasks_ffs12_percent_lags)
                                ffs21 += tasks_ffs21
                                ffs21_lags.extend(tasks_ffs21_lags)
                                ffs21_percent_lags.extend(tasks_ffs21_percent_lags)

                if fs12 + ffs12 + ss12 >= fs21 + ffs21 + ss21:
                    order_con = 1
                    fs = fs12
                    ffs = ffs12
                    ss = ss12
                else:
                    order_con = 2
                    fs = fs21
                    ffs = ffs21
                    ss = ss21

                if max([fs, ss, ffs]) != 0:
                    if fs > ss:
                        if ffs > 0:
                            if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'FFS',
                                                                      find_min_without_outliers(ffs12_percent_lags),
                                                                      count])
                            else:
                                predecessors_info_dict[w1_id].append([w2_id, 'FFS',
                                                                      find_min_without_outliers(ffs21_percent_lags),
                                                                      count])
                        else:
                            if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'FS', 0.0, count])
                            else:
                                predecessors_info_dict[w1_id].append([w2_id, 'FS', 0.0, count])
                    elif ss > ffs:
                        if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'SS',
                                                                      find_min_without_outliers(ss12_percent_lags),
                                                                      count])
                        else:
                            predecessors_info_dict[w1_id].append([w2_id, 'SS',
                                                                  find_min_without_outliers(ss21_percent_lags), count])
                    else:
                        if order_con == 1:
                                predecessors_info_dict[w2_id].append([w1_id, 'FFS',
                                                                      find_min_without_outliers(ffs12_percent_lags),
                                                                      count])
                        else:
                            predecessors_info_dict[w1_id].append([w2_id, 'FFS',
                                                                  find_min_without_outliers(ffs21_percent_lags), count])

    return predecessors_info_dict


def set_connections_info(graph_df: pd.DataFrame,
                         history_data: pd.DataFrame,
                         use_model_name: bool = False,
                         mapper: NameMapper | None = None,
                         change_connections_info: bool = False,
                         all_connections: bool = False,
                         id2ind: dict[str, int] = None) \
        -> pd.DataFrame:
    """
    Restore tasks' connection based on history data

    :param: change_connections_info - whether existing connections' information should be modified based on history data
    :param: expert_connections_info - whether existing connections should not be modified based on connection history data
    :return: repaired DataFrame
    """

    def update_connections(pred_ids_lst_tmp, pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst):
        """Updates existing connections with new data."""
        for i, pred_id in enumerate(pred_ids_lst_tmp):
            if pred_id in pred_ids_lst:
                idx = pred_ids_lst.index(pred_id)
                pred_types_lst_tmp[i] = pred_types_lst[idx]
                pred_lags_lst_tmp[i] = pred_lags_lst[idx]
                pred_counts_lst_tmp[i] = pred_counts_lst[idx]
        return pred_ids_lst_tmp, pred_types_lst_tmp, pred_lags_lst_tmp, pred_counts_lst_tmp

    def append_new_connections(pred_ids_lst_tmp, pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst):
        """Appends new connections to the existing lists."""
        for i, pred_id in enumerate(pred_ids_lst):
            if pred_id not in pred_ids_lst_tmp and pred_id != '-1' and pred_id in all_works:
                pred_ids_lst_tmp.append(pred_id)
                pred_types_lst_tmp.append(pred_types_lst[i])
                pred_lags_lst_tmp.append(pred_lags_lst[i])
                pred_counts_lst_tmp.append(pred_counts_lst[i])
        return pred_ids_lst_tmp, pred_types_lst_tmp, pred_lags_lst_tmp, pred_counts_lst_tmp

    tasks_df = graph_df.copy()

    INF = np.iinfo(np.int64).max

    # | ------ no changes ------- |
    if not change_connections_info and all_connections:
        tasks_df['counts'] = [[INF] * len(tasks_df['predecessor_ids'][i]) for i in range(tasks_df.shape[0])]

        tasks_df['connection_types'] = tasks_df['connection_types'].apply(
            lambda x: [EdgeType(elem) if elem != '-1' else EdgeType.FinishStart for elem in x]
        )
        return tasks_df

    # | ----------- for cache data ----------- |
    connections_dict = get_all_seq_statistic(history_data, graph_df, use_model_name, mapper)
    # with open('dormitory.json', 'w') as f:
    #     json.dump(connections_dict, f)

    all_works = tasks_df['activity_id'].values

    for task_id, pred_info_lst in connections_dict.items():
        if str(task_id) not in all_works:
            continue

        row = tasks_df.loc[id2ind[str(task_id)]]

        # | ------ change links and info about them ------ |
        # by default, the links and their information change
        if len(pred_info_lst) > 0:
            pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst = map(list, zip(*pred_info_lst))
        else:
            pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst = ['-1'], ['-1'], [1], [INF]

        if not all_connections:
            pred_ids_lst_tmp, pred_types_lst_tmp, pred_lags_lst_tmp, pred_counts_lst_tmp = (
                row['predecessor_ids'].copy(), row['connection_types'].copy(), row['lags'].copy(), [INF] * len(row['lags']))
            if change_connections_info:
                pred_ids_lst_tmp, pred_types_lst_tmp, pred_lags_lst_tmp, pred_counts_lst_tmp = (
                    update_connections(pred_ids_lst_tmp, pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst))
            pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst = (
                append_new_connections(pred_ids_lst_tmp, pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst))
        else:
            if row['predecessor_ids'] != ['-1']:
                # if 'lags' is unknown, thus 'connection_type' is also unknown
                pred_info_lst = [[pred_types_lst[i], pred_lags_lst[i], pred_counts_lst[i]]
                                 for i in range(len(pred_ids_lst)) if pred_ids_lst[i] in row['predecessor_ids']]
                pred_ids_lst = row['predecessor_ids']
                if len(pred_info_lst) > 0:
                    pred_types_lst, pred_lags_lst, pred_counts_lst = map(list, zip(*pred_info_lst))
                else:
                    pred_types_lst, pred_lags_lst, pred_counts_lst = ['-1'], [1], [INF]

        while len(pred_ids_lst) != len(pred_types_lst):
            pred_types_lst.append('FS')
            pred_lags_lst.append(1)
            pred_counts_lst.append(0)
        for col, val in zip(['predecessor_ids', 'connection_types', 'lags', 'counts'], [pred_ids_lst, pred_types_lst, pred_lags_lst, pred_counts_lst]):
            tasks_df.at[id2ind[str(task_id)], col] = val

    tasks_df['connection_types'] = tasks_df['connection_types'].apply(
        lambda x: [EdgeType(elem) if elem != '-1' else EdgeType.FinishStart for elem in x]
    )

    return tasks_df



import numpy as np

from sampo.api.genetic_api import FitnessFunction, ChromosomeType, ScheduleGenerationScheme
from sampo.base import SAMPO
from sampo.hybrid.population import PopulationScheduler
from sampo.scheduler.genetic import TimeFitness
from sampo.scheduler.genetic.schedule_builder import create_toolbox
from sampo.schemas import WorkGraph, Contractor, Time, LandscapeConfiguration, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec


class CycleHybridScheduler:
    def __init__(self,
                 starting_scheduler: PopulationScheduler,
                 cycle_schedulers: list[PopulationScheduler],
                 fitness: FitnessFunction = TimeFitness(),
                 max_plateau_size: int = 2):
        self._starting_scheduler = starting_scheduler
        self._cycle_schedulers = cycle_schedulers
        self._fitness = fitness
        self._max_plateau_size = max_plateau_size

    def _get_population_fitness(self, pop: list[ChromosomeType]):
        # return best chromosome's fitness
        return min(SAMPO.backend.compute_chromosomes(self._fitness, pop))

    def _get_best_individual(self, pop: list[ChromosomeType]) -> ChromosomeType:
        fitness = SAMPO.backend.compute_chromosomes(self._fitness, pop)
        return pop[np.argmin(fitness)]

    def run(self,
            wg: WorkGraph,
            contractors: list[Contractor],
            spec: ScheduleSpec = ScheduleSpec(),
            assigned_parent_time: Time = Time(0),
            landscape: LandscapeConfiguration = LandscapeConfiguration()) -> ChromosomeType:
        pop = self._starting_scheduler.schedule([], wg, contractors, spec, assigned_parent_time, landscape)

        cur_fitness = Time.inf().value
        plateau_steps = 0

        while True:
            pop_fitness = self._get_population_fitness(pop)
            if pop_fitness == cur_fitness:
                plateau_steps += 1
                if plateau_steps == self._max_plateau_size:
                    break
            else:
                plateau_steps = 0
                cur_fitness = pop_fitness

            for scheduler in self._cycle_schedulers:
                pop = scheduler.schedule(pop, wg, contractors, spec, assigned_parent_time, landscape)

        return self._get_best_individual(pop)

    def schedule(self,
                 wg: WorkGraph,
                 contractors: list[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 assigned_parent_time: Time = Time(0),
                 sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                 landscape: LandscapeConfiguration = LandscapeConfiguration()) -> Schedule:
        best_ind = self.run(wg, contractors, spec, assigned_parent_time, landscape)

        toolbox = create_toolbox(wg=wg, contractors=contractors, landscape=landscape,
                                 assigned_parent_time=assigned_parent_time, spec=spec,
                                 sgs_type=sgs_type)
        node2swork = toolbox.chromosome_to_schedule(best_ind)[0]

        return Schedule.from_scheduled_works(node2swork.values(), wg)
